import os
import enum
import cg3dguru.ui as ui
import pymel.core as pm
import cg3dguru.utils.modeling as mu


class Finger(enum.Enum):
    THUMB = 0,
    INDEX = 1,
    MIDDLE = 2,
    RING = 3,
    PINKY = 4,
    
    
class Joint(enum.Enum):
    METACARPAL = 0,
    PROXIMAL = 1,
    MEDIAL = 2,
    DISTAL = 3
    
    
JOINT_PERCENTS = {
    Finger.THUMB : {
        Joint.METACARPAL : -0.44,
        Joint.PROXIMAL : 0,
        Joint.DISTAL : 0.54
    }, 
    Finger.INDEX : {
        Joint.METACARPAL : -0.45,
        Joint.PROXIMAL : 0,
        Joint.MEDIAL : 0.49,
        Joint.DISTAL : 0.76
    },
    Finger.MIDDLE : {
        Joint.METACARPAL : -0.41,
        Joint.PROXIMAL : 0,
        Joint.MEDIAL : 0.48,
        Joint.DISTAL : 0.77
    },
    Finger.RING : {
        Joint.METACARPAL : -0.40,
        Joint.PROXIMAL : 0,
        Joint.MEDIAL : 0.47,
        Joint.DISTAL : 0.76
    },
    Finger.PINKY : {
        Joint.METACARPAL : -0.43,
        Joint.PROXIMAL : 0,
        Joint.MEDIAL : 0.46,
        Joint.DISTAL : 0.72
    }  
}



WINDOW_NAME = 'Plot Finger Percents'

class Fingers_Window(ui.Window):
    
    def __init__(self, windowKey, uiFilepath, *args, **kwargs):
        super(Fingers_Window, self).__init__(windowKey, uiFilepath)
        
        self.ui.thumb.pressed.connect( lambda: self.plot_joint(Finger.THUMB) )
        self.ui.index.pressed.connect( lambda: self.plot_joint(Finger.INDEX) )
        self.ui.middle.pressed.connect( lambda: self.plot_joint(Finger.MIDDLE) )
        self.ui.ring.pressed.connect( lambda: self.plot_joint(Finger.RING) )
        self.ui.pinky.pressed.connect( lambda: self.plot_joint(Finger.PINKY) )
        
        #I don't have support in for metacarpals
        self.ui.createMeta.setVisible(False)
        
                
        
    def plot_joint(self, finger : Finger):
        sel = pm.ls(sl = True)

        if len(sel) != 1:
            pm.system.displayError( 'One nurbsCurve transform must be selected.' )
            return

        shapes = pm.listRelatives(sel[0], shapes=True)
        if not shapes or not isinstance(shapes[0], pm.nodetypes.NurbsCurve):
            return
         
        curve = shapes[0]
        percents = JOINT_PERCENTS[finger]   
        for id in Joint:
            if id is Joint.METACARPAL and not self.ui.createMeta.isChecked():
                continue
            
            if id in percents:
                #create locator
                locator = pm.createNode('locator')
                marker = pm.listRelatives(locator, parent=True)[0] #get the shape's Transform
                
                name =  (finger.name + '_' + id.name).lower()
                pm.rename(marker, name)
                
                if id is Joint.METACARPAL:
                    mu.plot_percent_on_curve(curve, percents[id], marker)
                else:
                    mu.plot_percent_on_curve(curve, percents[id], marker)
                    
                    
        
        
def run():
    filepath = os.path.join( os.path.dirname(__file__), r'fingers.ui' )
    joint_utils_window = Fingers_Window(WINDOW_NAME, filepath)
    joint_utils_window.ui.show()


import pymel.core
import os
import re

#http://tech-artists.org/forum/showthread.php?4988-Problem-doing-an-FBX-export-with-PyMEL
#http://download.autodesk.com/global/docs/maya2014/en_us/index.html?url=files/GUID-377B0ACE-CEC8-4D13-81E9-E8C9425A8B6E.htm,topicNumber=d30e145135

_EXPORT_NODES = None

EXPORT_ANIM = 0x01
EXPORT_RIG  = 0x01 << 1
EXPORT_ANIM_RIG = EXPORT_ANIM | EXPORT_RIG

#Eventually get the save location from the node, if one doesn't exist add the attr to store the result
def get_save_filename(export_node=None):
    basicFilter = "*.fbx"
    filename = pymel.core.system.fileDialog2(fileFilter = basicFilter, cap='Export Animation')
    
    if filename:
        return filename[0]
    else:
        return None



def strip_namespace(filename):
    fbx = open(filename)
    file_string = fbx.read()
    fbx.close()
    
    results = re.finditer('(Model::[_:a-z0-9]*)', file_string, re.IGNORECASE)
    segments = []
    for i in results:
        start_end = ([i.start(), i.end()])
        segments.append(start_end)
        
    segments.reverse()
    for idx, i in enumerate(segments):
        model_name = file_string[i[0]:i[1]]
        model_name = model_name.replace('Model::', '')
        namespaces = model_name.split(':')
        model_name = 'Model::{0}'.format(namespaces[-1])
        
        pre_file  = file_string[0:i[0]]
        post_file = file_string[i[1]:]

        file_string = pre_file + model_name + post_file
        
    #save_name = filename.replace('.fbx', '_edited.fbx')

    new_file = open(filename, 'w')
    new_file.write(file_string)
    new_file.close()
    
    return None



def set_export_options(export_type):
    ##https://help.autodesk.com/view/MAYAUL/2022/ENU/index.html?guid=GUID-699CDF74-3D64-44B0-967E-7427DF800290
    start = int(pymel.core.animation.playbackOptions(query=True, animationStartTime=True))
    end   = int( pymel.core.animation.playbackOptions(query=True, animationEndTime = True))
    bake_anims = bool(export_type & EXPORT_ANIM)
    export_rig = bool(export_type & EXPORT_RIG)
    
    print('export animations:{0} export rig:{1}'.format(bake_anims, export_rig))
    print('start:{0} end:{1}'.format(start, end))
    
    pymel.core.mel.FBXResetExport()
    
    pymel.core.mel.FBXExportBakeComplexStart(v=start)
    pymel.core.mel.FBXExportBakeComplexEnd(v=end)
    pymel.core.mel.FBXExportBakeComplexAnimation(v= bake_anims )
    pymel.core.mel.FBXExportBakeResampleAnimation (v= True )
    pymel.core.mel.FBXExportSkins(v=export_rig )
    
    pymel.core.mel.FBXExportShapes(v=True)
    
    pymel.core.mel.FBXExportConstraints(v=False)
    pymel.core.mel.FBXExportInputConnections(v=False)
    #pymel.core.mel.FBXExportUseSceneName(v=True)   //This uses the maya filename for the clip being exported
    pymel.core.mel.FBXExportCameras(v=False)
    pymel.core.mel.FBXExportLights(v=False)
    
    pymel.core.mel.FBXExportInAscii(v=True)
    pymel.core.mel.FBXExportFileVersion(v='FBX201300')
    
    pymel.core.mel.FBXExportAnimationOnly(v=not export_rig)
    


def export(export_type=EXPORT_ANIM_RIG, filename='', strip_namespace=False):
    set_export_options(export_type)
    
    if not filename:
        filename = get_save_filename()
        
    if filename:
        pymel.core.mel.FBXExport(s=True, f=filename)
        if os.path.exists(filename) and strip_namespace:
            strip_namespace(filename)  
    

def export_anim(*args, **kwargs):
    export(export_type = EXPORT_ANIM_RIG, *args, **kwargs)
    
def export_rig(*args, **kwargs):
    export(export_type=EXPORT_RIG, *args, **kwargs)
    

def import_fbx(filepath):
    ##https://help.autodesk.com/view/MAYAUL/2022/ENU/index.html?guid=GUID-699CDF74-3D64-44B0-967E-7427DF800290
    pymel.core.mel.FBXImportMode(v='merge')
    pymel.core.mel.FBXImportFillTimeline(v=True)
    pymel.core.mel.FBXImportSkins(v=True)
    pymel.core.mel.FBXImport(f=filepath)


    


import sys
import os
import pymel.core as pm
import pkgutil
import importlib
from pathlib import Path

    
class MenuItem(object):
    """Base class for creating menu items, including menus and radio groups
    """
    def __init__(self, name, path, relative_path, is_menu):
        self.name = name
        self.parent = None
        self.children = {}
        self.path = path
        self.relative_path = relative_path
        
        #menu item info
        self.menu_instance = None
        self.divider = None
        self.params = {}
        self.order = []
        self.is_menu = is_menu
        self.is_radio_group = False
        self.is_options_box = False
        self.is_root_menu = False
        self.show = True
        
        #parsed functions
        self.command = None
        self.ddc_command = None
        self.dmc_command = None
        self.pmc_command = None
        
        
    def get_variable_name(self):
        item_name = 'MENU_ITEM'
        if self.is_options_box:
            item_name = 'OPTIONS_ITEM'
            
        return item_name
        
        
    def __set_args(self, module, attr_name, alt_name):
        """maps a function (if it exists) to its menu_item argument"""
        if not hasattr(module, attr_name):
            if attr_name in self.params:
                return self.params[attr_name]
            elif alt_name in self.params:
                return self.params[attr_name]
            else:
                return None
        
        attr = getattr(module, attr_name)
        
        if alt_name in self.params:
            self.params.pop(alt_name)
            
        self.params[attr_name] = attr
        
        return attr
        
        
    def parse_module(self):
        """Loads the python module associated with this object and parses it
        
        """
        #DONT try to parse on __init__, because this fuction creates an option box
        #with the same path info, which causes recursion.
        try:
            #only import if the modules isn't already imported
            module = sys.modules[self.relative_path]
        except:
            importlib.import_module(self.relative_path)
            module = sys.modules[self.relative_path]

        if hasattr(module, 'PARAMS'):
            self.params = module.PARAMS
        if hasattr(module, 'DIVIDER'):
            self.divider = module.DIVIDER
        if hasattr(module, 'ORDER'):
            self.order =  module.ORDER
        if hasattr(module, 'IS_RADIO_GROUP'):
            self.is_radio_group = module.IS_RADIO_GROUP
        if hasattr(module, 'SHOW'):
            self.show = module.show()
        if hasattr(module, 'OPTIONS'):
            options_name = "{}_{}".format(self.name, 'options')
            child_item =  MenuItem(options_name, self.path, self.relative_path, False)
            self.children[options_name] = child_item
            
            child_item.is_options_box = True
            child_item.params = module.OPTIONS
            if not 'optionBox' in child_item.params and not 'ob' in child_item.params:
                child_item.params['optionBox'] = True
                
                
        #lets finalize some of the kwargs
        self.command = self.__set_args(module, 'command', 'c')
        self.ddc_command = self.__set_args(module, 'dragDoubleClickCommand', 'ddc') 
        self.dmc_command = self.__set_args(module, 'dragMenuCommand', 'dmc')
        self.pmc_command =  self.__set_args(module, 'postMenuCommand', 'pmc') 
        
        
        if not 'label' in self.params and not 'l' in self.params:
            self.params['label'] = self.name
            
        key = 'label'
        if 'l' in self.params:
            key = 'l'
        self.params[key] =  self.params[key].replace("_", " ").capitalize()
        
        
        #flush any previous menu instances
        var_name = self.get_variable_name()
        menu_item = None
        if hasattr(module, var_name):
            menu_item = getattr(module, var_name)
            
        if hasattr(module, "IS_ROOT_MENU") and menu_item is not None:
            #If we're rebuiding this, we'll want to nuke the old object,
            try:
                print('deleting : {}'.format(menu_item.menu_instance))
                pm.deleteUI(menu_item.menu_instance)
                print('delete successful')
            except:
                pass
            
        setattr(module, var_name, None)
                

         
    def get_sorted_keys(self):
        """Returns the order to add menu items
        
        If the package contains a user defined order list in the __init__
        This is order is used, with any missing items added to the bottom
        of the list.
        
        If the user didn't define an order, then the order is based the module
        names.
        """
        sorted_children = list(self.children.keys())
        sorted_children.sort()
        if not self.order:   
            return sorted_children
        
        reordered = []
        for name in self.order:
            if name in sorted_children:
                sorted_children.pop(sorted_children.index(name))
                reordered.append(name)
                
        #Items might be missing from the order list, so let's add them to the end of the menu
        if sorted_children:
            item = self.children[sorted_children[0]]
            if not item.is_sub_menu: 
                item.divider = 'Needs Ordering'
            
        return reordered + sorted_children
            
    
    @property
    def is_sub_menu(self):
        return self.is_menu and not self.is_radio_group and self.divider is None
        
  
            
def _build_menu(menu_item, parent):
    """Walks the menu_item structure and builds the maya menu items"""                 
    if menu_item.is_root_menu:
        menu_item.params['parent'] = parent
        if 'p' in menu_item.params:
            menu_item.params.pop('p')
            
        if 'tearOff' not in menu_item.params and 'to' not in menu_item.params:    
            menu_item.params['tearOff'] = True
            
        menu_item.menu_instance = pm.menu(**menu_item.params)  
    else:
        if menu_item.divider is not None:
            pm.menuItem(divider = True, label = menu_item.divider)

        if menu_item.is_menu and not menu_item.is_radio_group and menu_item.divider is not None:
            #sub-folder with dividers don't get any menu item
            pass
        
        elif menu_item.is_radio_group:
            pm.radioMenuItemCollection()
            menu_item.menu_instance = parent
        else:
            if menu_item.is_sub_menu:
                menu_item.params['subMenu'] = True
            menu_item.menu_instance = pm.menuItem(**menu_item.params)
    
    
    ##I attempted to cache each module from MenuItem.parse_module() to a var called "menu_item.module"
    ##but the variable seems to lose scope(?totally bizarre), so I dropped the class variable and opted
    ##to make a local var instead.    
    module = sys.modules[menu_item.relative_path]   
    var_name = menu_item.get_variable_name()

    #Store the menu item in the module, so if we re-run this script we can
    #find the original menu_item.instance and delete it.
    setattr(module, var_name, menu_item)
    
    if menu_item.is_root_menu:
        setattr(module, 'IS_ROOT_MENU', True)        
    
        
    #Build the children items
    invalid_names = []
    for key in menu_item.get_sorted_keys():
        child_item =  menu_item.children[key]
        child_item.parent = menu_item
        if child_item.show:
            _build_menu(child_item, menu_item.menu_instance)
            
            if child_item.is_sub_menu:
                pm.setParent( '..', menu = True )
        else:
            invalid_names.append(key)
            
    #Let's clear out any objects that don't have menu items created for them
    for key in invalid_names:
        menu_item.children.pop(key)
        

    
def _walk_module(parent):
    """Creates a Menu Item Instance for each file/folder found under the parent path"""
    #https://stackoverflow.com/questions/1707709/list-all-the-modules-that-are-part-of-a-python-package
    for importer, modname, ispkg in pkgutil.iter_modules(path=[parent.path]):
        path = r'{}\{}'.format(parent.path, modname)
        relative_path = "{}.{}".format(parent.relative_path, modname)
        menu_item =  MenuItem(modname, path, relative_path, ispkg)
        parent.children[modname] = menu_item
        menu_item.parse_module()
            
    for key in parent.children:
        if parent.children[key].is_menu:
            _walk_module(parent.children[key])
             
        

def get_package_namespace(absolute_path):
    """returns the module name of the given py file."""
    def __Add_Parent_Module(name, path):
        found = False
        parent_module = os.path.join(path.parent, '__init__.py')
    
        if os.path.exists(parent_module):
            path = path.parent
            name = os.path.basename(path) + "." + name
            found = True
    
        return (found, name, path)    

    name = os.path.basename(absolute_path).split('.')[0]
    path = Path(absolute_path)
    loop = True
    count = 0
    while loop and count < 128:
        count += 1
        loop, name, path = __Add_Parent_Module(name, path)
    if count >= 128:
        pm.error('make menu found a bad module path:{}'.format(absolute_path))

    #special case for when someone's getting the path of an __init__ file
    if (name.endswith(".__init__")):
        name = name.removesuffix(".__init__")

    return name
         
        
    
    
def run(menu_bar = None, menu_namespace = ''):
    """Builds a menu based on a python package structure
    
    """
    if not menu_bar:
        menu_bar =  pm.language.melGlobals['gMainWindow']    
    
    if menu_namespace:
        if menu_namespace not in sys.modules:
            try:
                importlib.import_module(menu_namespace)  
            except:
                pm.error("Failed to find menu namespace")
            
        module = sys.modules[menu_namespace]
        path = module.__path__[0]
        
    else:
        try:
            from . import menu
            path = menu.__path__[0]
        except:
            pm.error("Failed to find menu namespace")

    menu_namespace = get_package_namespace(path)    

    package_name = menu_namespace.split('.')[0]
    main_menu = MenuItem(package_name, path, menu_namespace, True)
    main_menu.is_root_menu = True
    main_menu.parse_module()
    main_menu.is_root_menu = True
    
    _walk_module(main_menu)  
    _build_menu(main_menu, menu_bar)
    
    return main_menu


"""The udata module creates and manages Maya attributes in a pythonic way

Adding and managing custom attributes in Maya through Maya's standard low
level functions such as addAttr() has a few drawbacks that this higher-level
module attempts to overcome. These noteable improvements include:

1. udata handles all the back-end work of reading and writing
udata.BaseData class attributes to Maya attributes. Programmers and
Tech-artists can simply search for their Python classes inside of a Maya
scene or on a given node and work with the results.

2. Attributes in the attributes editor are organized under a compound
attribute, making it easy for end-users to understand how blocks of
attributes related to one another.

3. Built-in versioning support. Outdated attributes can easily be
identified and updated to match their Python equivalent as code evolves.

4. Less attribute flags and more automation. A number of addAttr() flags can
be determined automatically so you don't need to worry about things like -at
and -dt, parent or child count, or even making attributes for compounds where
the child size is predetermined.
"""

__author__ = "Nathaniel Albright"
__email__ = "developer@3dcg.guru"

VERSION = (0, 9, 0)
__version__ = '.'.join(map(str, VERSION))


import pymel.core as pm

#Don't change _RECORDS_NAME unless your project really desires an alternative
#name for the life of all scripts and tools that leverage the udata module
#(and remember to change it whenever getting an updated version of the module).
_RECORDS_NAME = 'DataRecords'
"""The name of the custom attr that tracks what data is on a node"""

DEFAULT_NODE_TYPE = 'network'
"""The nodeType that will be created when creating a node for storing data"""

AUTO_UPDATE = False
"""Should versioning attempt to auto update when there's a version mismatch?

Some studios may want outdated class data to automatically update. Setting
this to True will mean outdated data will attempt to update when it's
discovered. User's can decide this on a per-class instance by overriding
BaseData.pre_update_version().
"""

REPORT_WARNINGS = True
"""Should this module throw warnings?"""

class VersionUpdateException(Exception):
    """Thrown when BaseData.update_version() errors"""
    pass


class Attr(object):
    """A Wrapper for Maya's attribute arguements"""
    
    
    attr_types = set(['bool', 'long', 'short', 'byte', 'char', 'enum',
                      'float', 'double', 'doubleAngle', 'doubleLinear', 'message', 'time',
                      'fltMatrix'])
    """A list of attributeType names that use the -at flag"""    
    
    
    data_types = set(['string', 'stringArray', 'matrix',
                      'reflectanceRGB', 'spectrumRGB', 'float2', 'float3', 'double2',
                      'double3', 'long2', 'long3', 'short2', 'short3', 'doubleArray',\
                      'floatArray', 'Int32Array', 'vectorArray', 'nurbsCurve', 'nurbsSurface',\
                      'mesh', 'lattice', 'pointArray']
                     )
    """A list of attributeType names that use the -dt flag
    
    If an attr.attr_type is found in Attr.data_types then the 'dt' flag
    is automatically added to the args when creating the Maya attribute, else
    the 'at' flag is used.
    """
     
        
    def __init__(self, name, attr_type, *args, **kwargs):
        """The init func can take any arguments used in maya.cmds.addAttr()
                
        Users don't need to include the following flags:
        
        -longName or -ln : this is instead derived from the Attr.name.
        -attributeType or -at : this is determined by inspecting Attr.attr_type.
        -dataType or -dt : this is determined by inspecting Attr.attr_type.
        -parent or -p: Determined by the Compound class parent-child structure
        -numberOfChildren or -nc : Determined by the Compound class parent-child structure
        """
        super(Attr, self).__init__()
        
        self.name = name
        self.attr_type = attr_type
        self.args = args
                
        #combine any input flags with class defined flags
        flags = self.get_default_flags()
        flags.update( kwargs )
        self._clear_invalid_flags( flags )
        
        self._flags = flags
        

    @staticmethod
    def _clear_invalid_flags(flags):
        invalid = [ 'longName', 'ln', 'attributeType', 'at', 'dataType', 'dt', 'p', 'parent', 'numberOfchildren', 'nc']
        for key in invalid:
            if key in flags:
                flags.pop(key)
                
                
    @classmethod
    def get_default_flags(cls):
        """Defines class level flag arguments
        
        Sub-classes of Attr can override this function and return any
        flags used in maya.cmds.addAttr() so the class has a consistant
        look in Maya.        
        """
        return {}    
        
        
        
class Compound(Attr):
    """An attribute class that contains children attributes.
    
    For any -attributeType other than 'compound', users don't need to create
    the children attributes. For example: Compound("space", 'float3') will
    automatically create spaceX, spaceY, spaceZ.  If you don't want auto
    creation of child attributes then set make_elements=False on init.
    """
    
    compound_types = {'compound':0,
                     'reflectance':3, 'spectrum':3,
                     'float2':2, 'float3':3,
                     'double2':2, 'double3':3,
                     'long2':2, 'long3':3,
                     'short2':2, 'short3':3
                    }
    """A Dict of valid compound attr types and how many children to auto-create"""
    
    
    child_types = {'reflectance':'float', 'spectrum':'float',
                   'float2':'float', 'float3':'float', 'double2':'double',
                   'double3':'double', 'long2':'long', 'long3':'long', 'short2':'short',
                   'short3':'short'
                   }
    """What is the sub-element type that makes up this compound"""
    
    
    def __init__(self, name, attr_type, children = None, make_elements = True, *args, **kwargs):
        #reflectance and spectrum don't play nice with the color picker in the attr
        #so let's convert these to a float3 with a color flag
        if attr_type == 'reflectance' or attr_type == 'spectrum':
            attr_type = 'float3'
            kwargs['uac'] = True
                
        super(Compound, self).__init__(name, attr_type, *args, **kwargs)
        
        #don't make children = [] in the __init__ as it's persistant data
        #between class() calls.
        if children is None:
            children = []
        
        if attr_type not in Compound.compound_types:
            pm.error('udata Module: {0} is not a valid CompoundType.  Print Compound.CompoundTypes for valid list'.format(attr_type))
        
        self._target_size = Compound.compound_types[attr_type]
        self._children   = children
        
        if self._target_size and make_elements:
            if self._children:
                pm.error('udata Module: You supplied children, but also have make_elements = True.  Children will be erased')
            
            self._make_elements()

    
    def _make_elements(self):
        #define suffix for children
        xyz = ['X', 'Y', 'Z']
        rgb = ['R', 'G', 'B']
        
        suffix = xyz
        if 'usedAsColor' in self._flags or 'uac' in self._flags or \
           self.attr_type == 'spectrum' or self.attr_type == 'reflectance':
            suffix = rgb
            
        #determine child type
        type = Compound.child_types[self.attr_type]
            
        #add children to compound
        for i in range(0, self._target_size):
            attr = Attr(self.name + suffix[i], type)
            self.add_child(attr)
        
        
    def count(self):
        """How many children does this attribute have?"""
        return len(self._children)
    
    
    def add_child(self, child):
        """Add an attribute to this Compound as a child"""
        
        if self._target_size and len(self._children) > self._target_size:
            pm.error('udata Module: Compound instance has reached the max allowed children')
            
        self._children.append(child)
        
        
    def get_children(self):
        """Return the internal list of children"""
        return self._children
        
        
    def validate(self):
        """Confirm that the required number of children exist
        
        If the attr_type is 'compound' then one or more children must exist
        for this to be true. For all other attr_types the size of children
        must match the required count defined in Compound.compound_types
        """
        
        valid_size = False
        if not self._target_size:
            valid_size = len(self._children) > 0
        else:
            valid_size = self.count() == self._target_size
            
        if not valid_size:
            pm.error('udata module: {0} does not have the required number of children'.format(self.attr_type))



def create_attr(name, attr_type, *args, **kwargs):
    """A convience func returns an Attr() or Compound() based on the attr_type"""
    if attr_type in Compound.compound_types:
        return Compound(name, attr_type, *args, **kwargs)
    else:
        return Attr(name, attr_type, *args, **kwargs)
    


class Record(object):
    """The name of the class data found on a Maya node and its version info"""
    
    def __init__(self, attr):
        self._attr = attr
        self._name, str_version = attr.get().split(':')
        self._version = tuple(map(int, str_version.split('.')))
        

    @property
    def name(self):
        """The name of the class"""
        return self._name

        
    @property
    def version(self):
        """What version of the class is this record"""
        return self._version
    
    
    @version.setter
    def version(self, value):
        """Updates the version information for this record"""
        self.attr.unlock()
        self._version = value
        name = '{0}:{1}'.format( self.name, self._get_version_string() )
        self.attr.set( name )
        self.attr.lock()  
        
        
    @property
    def attr(self):
        """Returns the Maya attribute that represents this record"""
        return self._attr
    
    
    def _get_version_string(self):
        return '.'.join(map(str, self._version))
    
    
    
class BaseData(Attr):
    """Represents data that the user wants to store as Maya attributes
    
    Users should inherit from this class and at a minimum override
    get_atttributes(). get_attributes should either be declared as
    @classmethod or @staticmethod
    
    The python class names will become a compound attribute of the same name.
    Users can override cls.get_default_flags() to determine how this class
    looks as an attribute in Maya or pass the flags in on init().
    
    The BaseData class is responsible for reading and writing Maya attributes,
    creating and editing records, as well as class.version management.
    
    Any object that has a block of BaseData attributes stored on it will also
    contain a data block name that matches udata._RECORDS_NAME. Records are
    used for determing what version BaseData sub-classes are being stored on
    a given node. If a record for a given class is missing from the records
    then the udata module won't know that the data exists.
    """
    
    attributes = []
    """The attributes returned from get_atttributes()
    
    These attributes are stored at the class level so the data can be
    inspected and read without needing to create an instance of the class.
    Additionally, per instance attrs doesn't make sense in the scheme
    of how data is stored and used by this module.
    """

    _records = None
    _node = None
    _data_stack = []

    def __init__(self, *args, **kwargs):  
        super(BaseData, self).__init__(self.get_name(), 'compound', *args, **kwargs)
        
        self._init_class_attributes()

###----Versioning Methods----

    @classmethod
    def get_class_version(cls):
        return (0, 0, 0)
     

    @classmethod
    def get_version_string(cls):
        return '.'.join(map(str, cls.get_class_version()))

###----Record Methods-----

    @classmethod
    def _add_data_to_records(cls):
        if cls._records:

            indices = cls._records.getArrayIndices()
            if not indices:
                idx = 0
            else:
                idx = -1
                for i in range(0, indices[-1]):
                    if not i in indices:
                        idx = i
                        break
                    
                if idx == -1:
                    idx = indices[-1] + 1
            
            #concatenating the name and version is not as clean in code (vs
            #seperate attributes), but it makes end-user view from the
            #attribute editor clean while not taking up as much UI space
            name = '{0}:{1}'.format( cls.get_name(), cls.get_version_string() )
            cls._records[idx].set( name )
            cls._records[idx].lock()
            
                  
    @staticmethod
    def _create_records(node):
        global _RECORDS_NAME
        pm.addAttr(node, ln = _RECORDS_NAME,  dt = 'string', m= True)
                          
       
    @classmethod     
    def _get_records(cls, node, force_add ):
        if not node:
            pm.error('udata Module : Can\'t get records. node is None')
            
        has_records = pm.hasAttr(node, _RECORDS_NAME)
                
        if (not has_records) and force_add:
            cls._create_records( node )
            has_records = True
        
        if has_records:
            return node.attr(_RECORDS_NAME) #.records
        else:
            return None
        
 
    @classmethod
    def get_records(cls, node):
        """Sees if any records exists on the input node
        
        The records attributes will contain all list of all
        the records stored on the node.
        
        Returns the records attribute if found else None.
        """
        return cls._get_records(node, False)
 

    @classmethod
    def _get_record_by_name(cls, node, data_name):
        records = cls._get_records(node, False)
        found_record = None
        if records:
            for i in records.getArrayIndices():
                record = records[i]
                if not record.get():
                    continue
                
                name, version = record.get().split(':')
                if name == data_name:
                    found_record = Record(record)
                    break
                
        return found_record           

    
    @classmethod
    def get_record_by_name(cls, node, data_name):
        """Sees if a record by the input data_name exists on the input node
        
        Returns the record if it exists, else None        
        """
        return cls._get_record_by_name( node, data_name )
    
    
    @classmethod
    def get_record(cls, node):
        """see if a record matching the get_name() on the input node
        
        Returns the record if it exists, else None         
        """
        return cls._get_record_by_name( node, cls.get_name() )
    
                  
###----Attribute Methods----
 
    @classmethod
    def _find_attr_conflicts(cls):
        attr_names = cls.get_attribute_names()
        flags = cls.get_default_flags()
        cls._clear_invalid_flags(flags)
        
        if 'multi' in flags or 'm' in flags:
        #if 'multi' in cls._flags or 'm' in cls._flags:
            #I *believe* attributes that are part of a multi arg won't conflict.            
            attr_names = []       
        
        attr_names.append( cls.get_name() )
        
        conflicts = []
        for attr_name in attr_names:
            if pm.hasAttr(cls._node, attr_name):
                conflicts.append(attr_name)
                
        if conflicts:
            record_names = []
            for i in cls._records.getArrayIndices():
                record = Record(cls._records[i])
                record_names.append( record.name )            

            class_name = cls.__name__
            errorMessage = 'udata Attribute Conflict :: Attribute Name(s) : {0} from class "{1}" conflicts with one of these existing blocks of data : {2}'
            pm.error( errorMessage.format(conflicts, class_name, record_names) )
 
    @classmethod          
    def _add_attr(cls, attr, prefix, parent_name):
        attr_name = prefix + attr.name
        
        if parent_name:
            attr._flags['parent'] = parent_name
           
        if isinstance(attr, Compound):
            attr.validate()
            pm.addAttr(cls._node, ln = attr_name, at = attr.attr_type, nc = attr.count(), **attr._flags)
            for child in attr.get_children():
                cls._add_attr(child, '', attr_name)
            
        elif attr.attr_type in Attr.data_types:
            pm.addAttr(cls._node, ln = attr_name, dt = attr.attr_type, **attr._flags)             
        else:   
            pm.addAttr(cls._node, ln = attr_name, at = attr.attr_type, **attr._flags)     
           
           
    @classmethod                   
    def _get_attribute_names(cls, attr, name_list):
        name_list.append(attr.name)
        if isinstance(attr, Compound):
            for child in attr.get_children():
                cls._get_attribute_names(child, name_list)
            
            
    @classmethod        
    def get_attribute_names(cls):
        """returns a flat list of all the attribute names in class.attributes"""
        name_list = []
        cls._init_class_attributes()
            
        for attr in cls.attributes:
            cls._get_attribute_names(attr, name_list)
            
        return name_list

      
    @classmethod
    def _init_class_attributes(cls):
        if not cls.attributes:
            cls.attributes = cls.get_attributes()
        
        
    @classmethod
    def get_attributes(cls):
        """MUST IMPLEMENT : A list of udata.Attrs stored in this class.
        
        Sub-classes of BaseData must override this function and return
        the attributes they want to create and store in Maya. The function
        must also be declared as a class or static method.        
        """
        pm.error( 'udata Module: You\'re attempting to get attributes for class {0} that has no get_attributes() overridden'.format(cls.__name__) )       
       
       
###----Update Methods----
    
    @classmethod
    def pre_update_version(cls, old_data, old_version_number):
        """Determines if the update_version() should be called.
        
        Users can override this function if they wish to define their own
        logic. By default the function returns udata.AUTO_UPDATE, which
        is False by default.
        
        Args:
            old_data (pymel.general.Attr) : The data that's about to be updated.
            old_version_number (Tuple) : The version information of the old_data.
        
        Returns:
            Bool: True if update_version() should be called, else False.
        """
        global AUTO_UPDATE
        if not AUTO_UPDATE and REPORT_WARNINGS:
            pm.warning('cg3dguru.udata : Skip updating old data on "{}" of type "{}" because AUTO_UPDATE is False. Consider overriding pre_update_version(). Repress this warning with udata.REPORT_WARNINGS = False'
                       .format(old_data.node(), cls.__name__))
        
        return AUTO_UPDATE
     
    @classmethod  
    def update_version(cls, old_data, old_version_number):
        """Updates the data to the latest version of the Python defintion.
        
        The default implimentation of this happens in five steps.
        1. A temporary node is created with the latest class data
        2. The current data values are copied to the temp node using pymel.core.copyAttr
        3. The current data is deleted off the original node
        4. The new data is added to the original node.
        5. The data values are transferred back from the temp node to the original node.
        
        Failure of this process could occur during step #2. In this situation
        udata.VersionUpdateException is raised and the user will need to
        determine their own logic for how to update/replace their existing
        data with the new class version.
        
        Args:
            old_data (pymel.general.Attr) : the data of the outdated version.
            old_version_number (tuple) : The Max, min, patch value of the old data.
        
        Returns:
            Bool : True if the update was successful else False.
        """
        
        
        #Copy the attribute values to a temporary node
        #cls._node = old_data.node()
        
        temp_node, data = cls.create_node(name = 'TRASH', ss=True)
        name_list = cls.get_attribute_names()
        try:
            pm.copyAttr(cls._node, temp_node, at=name_list, ic=True, oc=True, values=True)
        except:
            #delete the tempNode
            pm.delete(temp_node)
            
            message = 'Please impliment custom update logic for class: {0}  oldVersion: {1}  newVersion: {2}'.format( cls.get_name(), old_version_number, cls.get_class_version())
            raise VersionUpdateException(message)
        
        #delete the attributes off the current node
        pm.deleteAttr( cls._node, at = cls.get_name() )
        
        #rebuild with the latest definition
        cls._create_data()

        #transfer attributes back to original node
        pm.copyAttr(temp_node, cls._node, at = name_list, ic = True, oc = True, values = True)  
        
        #delete the tempNode
        pm.delete(temp_node)
        
        return True

    @classmethod    
    def post_update_version(cls, data, update_successful):
        """Called after update_version()
        
        The default implimentation does nothing. Users can override
        this function if they want to do any post processing after
        the update_function has been run.
        
        args:
            data (pymel.general.Attr) : The newly updated data if successful
            else the old data.
            update_successful (bool) : Was the update successful.
        """        
        pass
    
    
###----Data Methods----
    
    @classmethod
    def get_name(cls):
        """The name of the class as it will appear in Maya's extra attributes
        
        By default the name is the same as the class.__name__ This should be
        determined upfront before using the name in production.  Once changed
        any existing records and associated data will be orphaned.  If users
        opt to override this, they must include a @classmethod declarator
        """
        return cls.__name__
    
    
    @classmethod
    def get_prefix(cls):
        """Retuns the prefix to append to all class attributes
        
        The deault prefix is an empty string. The prefix is designed to limit
        the potential of an attribute name collision between mutliple
        udata blocks assigned to the same maya node. Users can return ''
        if they don't want a prefix added to their attribute names. Any valid
        prefix will automaticallly be separated by '_'
        
        NOTE: If you decided to change the prefix once the data is production
        then you'll need to override update_version() with your own logic
        as old names and new names won't match when updating the data block.        
        """
        return ''
    
    
    @classmethod
    def get_attr_name(cls, attr_name):
        """return the input attribute name with an class prefix"""
        prefix = cls.get_prefix()
        if prefix:
            prefix += '_'
            
        return prefix + attr_name

            
    @classmethod
    def _create_data(cls):     
        attrs = cls.attributes
        long_name = cls.get_name()
        
        flags = cls.get_default_flags()
        cls._clear_invalid_flags(flags)        
        
        pm.addAttr(cls._node, ln = long_name, at = 'compound', nc = len( attrs ), **flags )

        prefix = cls.get_prefix()
        if prefix:
            prefix += '_'

        for attr in attrs:
            cls._add_attr(attr, prefix, long_name)       
         
        return cls._node.attr(long_name)
        
    @classmethod   
    def post_create(cls, data):
        """Called after the data has been created.
        
        This function doesn't do anything by default and exists purely
        for end-user convenience.  What a great place for adding your new
        data to an existing node network!
        
        Args:
            data (pymel.general.Attr) : The newly created data. 
        
        """        
        pass
              
    @classmethod
    def get_data(cls, node, force_add = False):
        """Attempts to return the class data stored on the input node.
        
        Args:
            node (pyNode) : The node you want to get/store data on
            force_add (bool) : Should the data be added if none exists?
            
        Returns:
            pymel.general.Attr : The data stored on the node or None.        
        """

        #We need to cache the current data, because get_data might be called
        #while doing an update. As this function gets deeper in scope, we
        #need to make sure we reset the data as we leave the scope. reseting
        #happens on exit.
        cls._data_stack.append((cls._records, cls._node))
        
        cls._records = None
        cls._node    = node
        
        record = cls.get_record(node)
        
        if not record and force_add:
            #lets make sure the records data exists
            cls._records = cls._get_records(node, force_add = True)
            
        #If found make sure the data block doesn't need updating.
        if record:
            data_name = cls.get_name() 
            record_version = record.version
            current_version = cls.get_class_version()
            
            #Attempt to updat the version
            if record_version < current_version:
                old_data = cls._node.attr(data_name)
                
                if cls.pre_update_version(old_data, record_version):
                    updated = cls.update_version(old_data, record_version)
                    if updated:
                        record.version = current_version
                                
                    data = cls._node.attr(data_name)
                    cls.post_update_version( data, updated )

            
            data = cls._node.attr(data_name)
                    
        #else, add the data to the node           
        elif force_add:
            cls._find_attr_conflicts()
            data = cls._create_data()
            cls._add_data_to_records()
            cls.post_create( data )
            
        else:
            data = None

        cls._records, cls._node = cls._data_stack.pop(-1)

        return data      
    
    @classmethod  
    def add_data(cls, node):
        """Add class attributes to the input node.
        
        If the input node already has data on it then versioning is run to
        ensure the data format is up-to-date.
        
        Args:
            node (pyNode) : The node to add the data to.
                
        Returns:
            pymel.general.Attr : The data added to the node.
        """
        return cls.get_data(node, force_add=True)
    

    @classmethod
    def delete_data(cls, node):
        """Remove the class attributes off of the input node

        Args:
            node (pyNode) : The node to remove the data from.        
        """
        record = cls.get_record(node)
        
        if record:
            record.attr.unlock()
            pm.removeMultiInstance(record.attr, b=True)
            pm.deleteAttr(node, at = cls.get_name() )   
            

###----Misc Methods----
    
    @classmethod
    def create_node(cls, nodeType = DEFAULT_NODE_TYPE, *args, **kwargs):
        """Creates a new node in Maya that also contains the Class attributes.
        
        The default node type that's created is driven by
        udata._DEFAULT_NODE_TYPE, but users can override this based on
        the input params.
        
        Args:
            nodeType (str, optional) : The name of the maya node to create.
            
        Returns:
            Tuple : The newly created pyNode as element zero and the data as
            element 1.
        """
        pynode = pm.general.createNode(nodeType, **kwargs)
    
        if pynode:
            #classInstance = cls()
            data = cls.get_data(pynode, force_add=True)
            
        return (pynode, data)   
    

    
class Utils(object):
    """Easy module and maya scene inspection
    
    The udata.Utils class provides a number of functions
    to inspect not only the active Maya scene, but also an inspection
    of Python classes that derive from udata.BaseData. Users
    can verify that there won't be any attribute naming conflicts between
    the various class defintions, search for data that's being used
    in the active scene, as well as verify verioning.    
    """
    def __init__(self, *args, **kwargs):
        super(Utils, self).__init__(*args, **kwargs)
    
    
    @staticmethod
    def get_classes():
        """Returns all BaseData subclass"""
        classes = BaseData.__subclasses__()
        return classes
    
    
    @staticmethod
    def get_class_names():
        """Returns a list of all BaseData subclass names."""
        sub_classes = Utils.get_classes()
        class_names = {}
        for subclass in sub_classes:
            class_names[ subclass.get_name() ] = subclass
            
        return class_names

    
    @staticmethod
    def find_attribute_conflicts(error_on_conflict = True):
        """Find any attribute naming conflicts between the BasData subclasses.
        
        Since the BaseData.attributes are written to Maya as basic attributes
        it's possible to have a naming conflict between to classes. For
        example myclass.startFrame and exportData.startFrame. Users can call
        this function after adding any new class definition to ensure there
        aren't any naming conflicts. In reality, if two classes do have a
        naming conflict but the production pipeline means those two blocks of
        data would never be stored on the same object, then the conflict
        should be a non-issue. this function will print any found conflicts
        in the output window and optionally raise an error if a conflict has
        been found.
        
        Args:
            error_on_conflict (bool) : Should an error be raised if a
            conflict is found?
        """
        
        conflicts = {}
        attrs = {}
        for subclass in Utils.get_classes():
            default_flags = subclass.get_default_flags()
            
            #if the Class is going to be added with a mutli flag, then
            #there shouldn't be any conflicts with its attributes
            if 'm' in default_flags or 'multi' in default_flags:
                continue
            
            attribute_names = subclass.get_attribute_names()
            for attr_name in attribute_names:
                if attr_name not in attrs:
                    attrs[attr_name] = [subclass.__name__]
                else:
                    attrs[attr_name].append(subclass.__name__)
                    
        for attr_name in attrs:
            classes = attrs[attr_name]
            
            if len(classes) > 1:
                conflicts[attr_name] = classes
                
        if conflicts:
            for attr_name in conflicts:
                classes = conflicts[attr_name]
                if REPORT_WARNINGS:
                    pm.warning( 'udata.Utils :: Attr conflict. : "{0}" exists in classes: {1}'.format(attr_name, classes))
                
            if error_on_conflict:
                pm.error( 'udata.Utils :: Found conflict between attribute names. See console for info.')
                
        return conflicts
    
    
    @staticmethod
    def get_nodes_with_data(nodes = None, data_class = None, *args, **kwargs):
        """Return a list of nodes that have any (or specific) data attached
        
        Users can pass in a list of nodes to examine or if no nodes are
        provided they can pass in the standard pymel.core.ls args to generate
        a list of node to inspect.
        
        Args:
            nodes (pyNode list, optional) : What nodes should the function
            search?
            data_class (BaseData sub-class, optional) : What class data
            are we looking for. If this is None then a node will be
            returned if it has any data attached to it.
            **kwargs (pymel.ls flags) : Only considered if node is None.
            
        Returns:
            list : A list of pyNodes that match the given search criteria. 
        """
        if not nodes:
            nodes = pm.ls(*args, **kwargs)
            nodes.sort()
        
        data_nodes = []
        for node in nodes:
            if data_class:
                records = BaseData.get_record_by_name( node, data_class.get_name() )
            else:
                records = BaseData.get_records(node)
                
            if records:
                data_nodes.append(node)
                
        return data_nodes
    
    
    @staticmethod
    def validate_version(nodes = None, *args, **kwargs):
        """Force version validation on the given node conditions

        This function can be used to ensure current scene nodes have their
        data up-to-date. Users can pass in a list of nodes to examine or if
        no nodes are provided they can pass in the standard pymel.core.ls
        args to generate a list of node to inspect.
        
        Validating a scene might look something like this:
        
        #find all scene nodes that have any data on them
        nodes = Utils.get_nodes_with_data()
        
        #Update the nodes we found in our scene.
        Utils.validate_version(nodes = nodes)
        
        If you're feeling really lazy you can just examine every object in
        the scene against every potential class like this:
        
        Utils.validate_version()
            
        Args:
            nodes (pyNode list, optional) : What nodes should the function
            search?
            **kwargs (pymel.ls flags) : Only considered if nodes is None.
        """
        if not nodes:
            nodes = pm.ls(*args, **kwargs)        
                
        classes= Utils.get_classes()
        
        for data_class in classes:
            #instance = data_class()
            for node in nodes:
                #this forces a version check
                data_class.get_data(node)
                #instance.get_data(node)

        
            
    
    





## how to get QT Designer
## https://stackoverflow.com/questions/30222572/how-to-install-qtdesigner
## https://build-system.fman.io/qt-designer-download
## Designer is already part of Maya (check your /bin folder)


import os

from maya import OpenMayaUI as omui 

from PySide2.QtCore import * 
from PySide2.QtWidgets import *
from PySide2.QtUiTools import *
from PySide2 import __version__
from shiboken2 import wrapInstance 


class Window(object): #QWidget):
    CUSTOM_WINDOWS = {}
    mayaWindow = None
    
    
    def __init__(self, windowKey, uiFilepath, custom_widgets = None, *args, **kwargs):
        if not os.path.exists(uiFilepath):
            raise FileNotFoundError("Invalid path {}".format(uiFilepath))
        
        self.add_window(windowKey, self)
        
        loader = QUiLoader()
        if custom_widgets:
            for widget in custom_widgets:
                loader.registerCustomWidget(widget)
        
        self.mainWindow = self.get_maya_window()
        self.ui = loader.load( uiFilepath, parentWidget = self.mainWindow )   
  
    @classmethod
    def add_window(cls, windowKey, instance):
        if windowKey in cls.CUSTOM_WINDOWS:
            try:
                cls.CUSTOM_WINDOWS[windowKey].ui.deleteLater()     
            except:
                pass
            
        cls.CUSTOM_WINDOWS[windowKey] = instance
        
   
    @classmethod
    def get_maya_window(cls):
        if not cls.mayaWindow:
            _mayaMainWindowPtr = omui.MQtUtil.mainWindow()
            cls.mayaWindow     = wrapInstance( int(_mayaMainWindowPtr), QMainWindow) 
            
        return cls.mayaWindow
    
    
    
#def Run():

    #filepath = os.path.join( os.path.dirname(__file__), r'ui\UserData.ui' )
    #userData = Window('test', filepath)
    #userData.ui.show()    



import os
import sys
import importlib

import pymel.core as pm

import cg3dguru.ui as ui
import cg3dguru.udata

WINDOW_NAME = 'User Data Editor'

       
class UserDataEditor(ui.Window):
    
    def __init__(self, windowKey, uiFilepath, *args, **kwargs):
        super(UserDataEditor, self).__init__(windowKey, uiFilepath)

        self.add_script_job()
        self.maya_nodes_selected = len(pm.ls(sl=True)) > 0

        self.classes = cg3dguru.udata.Utils.get_class_names()
        keys = list(self.classes.keys())
        keys.sort()
        self.ui.createDataList.addItems(keys)
        self.ui.searchDataList.addItems(keys)
        
        self.ui.createDataList.itemSelectionChanged.connect(lambda: self.on_selection_changed(self.ui.createDataList))
        self.ui.searchDataList.itemSelectionChanged.connect( lambda : self.on_selection_changed(self.ui.searchDataList) )
        
        self.ui.createData.clicked.connect(self.on_create)
        self.ui.addData.clicked.connect(self.on_add)
        self.ui.removeData.clicked.connect(self.on_delete)
        self.ui.sceneSelect.clicked.connect(self.on_select_from_scene)
        self.ui.filterSelection.clicked.connect(self.on_find_in_selection)
        self.ui.attribute_conflicts.clicked.connect(self.on_attribute_conflicts)
        
        
    def on_attribute_conflicts(self, *args, **kwargs):
        conflicts = cg3dguru.udata.Utils.find_attribute_conflicts(error_on_conflict=False)

        output = ''
        if conflicts:
            for attr_name in conflicts:
                classes = conflicts[attr_name]
                output += 'Attr Conflict: "{0}" exists in classes: {1}\n'.format(attr_name, classes)
        else:
            output = 'No attribute conflicts found. :)'

        self.ui.report_results.clear()
        self.ui.report_results.setPlainText(output)
        
        
    def add_script_job(self):
        jobId   = pm.scriptJob( event=['SelectionChanged', self.maya_selection_changed] )
        #print 'New Job: {0}'.format(jobId)
        self.handler = lambda : self.remove_script_job(jobId)
        #self.jobId = jobId
        self.ui.destroyed.connect( self.handler )        
        
        
    def remove_script_job(self, jobId):
        #print 'Nuke Job: {0}'.format(jobId)
        self.ui.destroyed.disconnect( self.handler )
        pm.scriptJob( kill = jobId )
        
        
    def _get_item_names(self, listWidget):
        selection = listWidget.selectedItems()
        names = []
        for item in selection:
            names.append( item.text() )
            
        return names
        
        
    def on_create(self):
        names = self._get_item_names(self.ui.createDataList)
        newNodes = []
        
        for name in names:
            data_class = self.classes[name]
            pyNode, data = data_class.create_node(name = name)
            newNodes.append(pyNode)
                
        if newNodes:
            pm.select(newNodes, replace = True)
        else:
            self.ui.statusbar.showMessage("No Data is selected")
    
    
    def on_add(self):
        names = self._get_item_names(self.ui.createDataList)
        cg3dguru.udata.Utils.validate_version(sl=True)
        
        for name in names:
            data_class = self.classes[name] #()
            for maya_node in pm.ls(sl=True):
                data_class.add_data( maya_node )
                
        self.on_selection_changed(self.ui.createDataList)
    
    
    def on_delete(self): 
        selection = pm.ls(sl=True)
        names = self._get_item_names(self.ui.createDataList)
        
        for name in names:
            data_class = self.classes[name] #()
            for mayaNode in selection:
                data_class.delete_data( mayaNode )
                
        self.maya_selection_changed()
    
    
    
    def _select(self, *args, **kwargs):
        names = self._get_item_names(self.ui.searchDataList)
        
        nodes = []
        for name in names:
            data_class = self.classes[name] #()
            found_nodes = cg3dguru.udata.Utils.get_nodes_with_data(data_class=data_class, **kwargs)
            
            nodes.extend(found_nodes)
            
        pm.select(nodes, replace=True)
        if not nodes:
            self.ui.statusbar.showMessage('No data found!', 5000)
        else:
            self.ui.statusbar.clearMessage()



    def on_select_from_scene(self):
        self._select()
            
    
    def on_find_in_selection(self):
        self._select(sl=True)
    
    
    def on_selection_changed(self, list_widget):
        #is anything selected in our list?
        enable = len( list_widget.selectedItems() ) > 0
        
        if list_widget is self.ui.createDataList:
            names = self._get_item_names(self.ui.createDataList)
            hasData  = False
            missData = False
            
            for name in names:
                data = None
                data = cg3dguru.udata.Utils.get_nodes_with_data(data_class = self.classes[name], sl=True)
                if data:
                    hasData  = True
                
                missData = len(data) != len(pm.ls(sl=True))
            
     
            self.ui.createData.setEnabled(enable)   
            self.ui.addData.setEnabled(enable and self.maya_nodes_selected and missData)
            self.ui.removeData.setEnabled(enable and hasData)
        else:
            print ("search list")
            
            
    def maya_selection_changed(self):
        self.maya_nodes_selected = len( pm.ls(sl=True) ) > 0
        
        if self.ui.createDataList.isVisible():
            self.on_selection_changed(self.ui.createDataList)
        else:
            self.on_selection_changed(self.ui.searchDataList)
    
 
def run(data_module = None):
    if data_module is not None:
        if data_module not in sys.modules:
            try:
                importlib.import_module(data_module)
            except Exception as e:
                print("failed to import {}".format(data_module))
                return

    filepath = os.path.join(cg3dguru.udata.__path__[0],  'user_data.ui' )
    editor = UserDataEditor(WINDOW_NAME, filepath)
    editor.ui.show()
    


import math
import enum
import maya.cmds
import pymel.core as pm


class Axis(enum.Flag):
    X = enum.auto()
    Y = enum.auto()
    Z = enum.auto()
    ALL = X | Y | Z
    REVERSE = enum.auto()
    NEG_X = X | REVERSE
    NEG_Y = Y | REVERSE
    NEG_Z = Z | REVERSE
 
 
class Space(enum.Enum):
    OBJECT = 0
    WORLD = 1
    PARENT = 2
    CAMERA = 3
    TANGENT = 4
    UV = 5
    
    
class Direction(enum.Flag):
    FORWARD = enum.auto()
    UP = enum.auto()
    RIGHT = enum.auto()
    REVERSE = enum.auto()
    BACKWARDS = FORWARD | REVERSE
    DOWN = UP | REVERSE
    LEFT = RIGHT | REVERSE
    
    
class Flip(enum.Enum):
    FORWARD = 0
    UP = 1
    RIGHT = 2
    
    
class MatrixUtils(object):
    """
    Allows users to get and set transforms in the generic terms of forward, up, and right (fur).
    Users define the mapping of a axis to a direction.
    """
    def __init__(self, transform_node = None, *args, **kwargs):  
        super().__init__(*args, **kwargs)
            
        self._transform_node = None
        self.transform_node = transform_node
        
        self._right_axis = Axis.X
        self._up_axis = Axis.Y
        self._forward_axis = Axis.Z
        
        self._flip = Flip.RIGHT
        
        #what space the matrix operations take place in
        self._space = Space.WORLD
        

##--properties----------------------------------------------------


    @property
    def transform_node(self) -> pm.nodetypes.Transform:
        return self._transform_node
    
    @transform_node.setter
    def transform_node(self, node: pm.nodetypes.Transform):
        if node is not None and not isinstance( node, pm.nodetypes.Transform ):
            pm.system.error('Type Error')
            
        self._transform_node = node        
    
    @property
    def space(self) -> Space:
        return self._space
    
    @space.setter
    def space(self, value: Space):
        if not isinstance(value, Space):
            pm.system.error('Type Error')
        
        self._space = value;
        
    @property
    def Flip(self) -> Flip:
        return self._flip
    
    @Flip.setter
    def Flip(self, value: Flip):
        if not isinstance(value, Flip):
            pm.system.error('Type Error')
            
        self._flip = value
    
    @property
    def forward(self) -> pm.datatypes.Vector:
        return MatrixUtils.get_axis_vector( self.get_matrix(), self._forward_axis)
     
    @property
    def up(self) -> pm.datatypes.Vector:
        return MatrixUtils.get_axis_vector( self.get_matrix(), self._up_axis)
    
    @property
    def right(self) -> pm.datatypes.Vector:
        return MatrixUtils.get_axis_vector( self.get_matrix(), self._right_axis)    
    
 
##--static methods----------------------------------------------------------------------------
     
    @staticmethod
    def get_world_pos( obj_name ):
        x_list = maya.cmds.xform( obj_name, query = True, matrix = True, worldSpace = True )
        return pm.datatypes.Vector( *x_list[12:15] )    

    @staticmethod
    def get_matrix_position(matrix: pm.datatypes.Matrix):
        return pm.datatypes.Vector(matrix.a30, matrix.a31, matrix.a32)
    
    
    @staticmethod
    def set_matrix_translation(matrix: pm.datatypes.Matrix,
                        position: pm.datatypes.Vector):
        matrix.a30 = position.x
        matrix.a31 = position.y
        matrix.a32 = position.z
    
    
    @staticmethod
    def get_axis_vector(matrix: pm.datatypes.Matrix,
                        axis: Axis):
        
        vector = None
        if Axis.X in axis:
            vector = pm.datatypes.Vector(matrix.a00, matrix.a01, matrix.a02)
        elif Axis.Y in axis:
            vector = pm.datatypes.Vector(matrix.a10, matrix.a11, matrix.a12)
        elif Axis.Z in axis:
            vector = pm.datatypes.Vector(matrix.a20, matrix.a21, matrix.a22)
            
        if Axis.REVERSE in axis:
            vector = -vector
            
        return vector
        
        
    @staticmethod
    def set_axis_vector(matrix: pm.datatypes.Matrix,
                        vector: pm.datatypes.Vector,
                        axis: Axis):
        if axis == Axis.X:
            matrix.a00 = vector.x
            matrix.a01 = vector.y
            matrix.a02 = vector.z
        if axis == Axis.Y:
            matrix.a10 = vector.x
            matrix.a11 = vector.y
            matrix.a12 = vector.z
        if axis == Axis.Z:
            matrix.a20 = vector.x
            matrix.a21 = vector.y
            matrix.a22 = vector.z
            
            
    @staticmethod
    def ensure_right_handedness( forward, up, right, handedness_rule: Flip):
        if not (right.cross( up ) * forward > 0 and up.cross( forward ) * right > 0):
            if handedness_rule == Flip.RIGHT:
                right.x = -right.x
                right.y = -right.y
                right.z = -right.z
            elif handedness_rule == Flip.UP:
                up.x = -up.x
                up.y = -up.y
                up.z = -up.z
            elif handedness_rule == Flip.FORWARD:
                forward.x = -forward.x
                forward.y = -forward.y
                forward.z = -forward.z    
            
            
    @staticmethod
    def set_matrix_vectors(matrix: pm.datatypes.Matrix, 
                           x: pm.datatypes.Vector,
                           y: pm.datatypes.Vector,
                           z: pm.datatypes.Vector,
                           position: pm.datatypes.Vector, 
                           flip: Flip,
                           ignore_scale: bool):
        """
        Set's the transform_node's f, u, r to the input while ensuring proper right handedness.
        """
        
        MatrixUtils.ensure_right_handedness(z, y, x, flip)
        
        if ignore_scale:
            z.normalize()
            y.normalize()
            x.normalize()
        
        MatrixUtils.set_axis_vector(matrix, x, Axis.X)
        MatrixUtils.set_axis_vector(matrix, y, Axis.Y)
        MatrixUtils.set_axis_vector(matrix, z, Axis.Z)
        MatrixUtils.set_matrix_translation(matrix, position)
        
        
    @staticmethod    
    def get_three_point_matrix( p1, p2, p3, u_dir = None ):
        f_vec = p3 - p1
        f_vec.normalize()
        
        r_vec = p2 - p1
        r_vec.normalize()

        if math.fabs( f_vec * r_vec ) > .999:
            maya.OpenMaya.MGlobal.displayError( "Your three points are too in line with one another!" )
            return None

        r_vec, u_vec = MatrixUtils.get_orthogonal_vectors( f_vec, r_vec, v3_dir = u_dir )
        #self.ensure_right_handedness( r_vec, u_vec, f_vec )

        matrix = pm.datatypes.Matrix()
        MatrixUtils.set_matrix_vectors(matrix, f_vec, u_vec, r_vec, p1, Flip.RIGHT, True)

        #pos = cgkit.cgtypes.vec4( p1 )
        #pos.w = 1

        #three_point_matrix = cgkit.cgtypes.mat4( )
        #three_point_matrix.setColumn( 3, pos )

        #orient = cgkit.cgtypes.mat3( r_vec, u_vec, f_vec )
        #three_point_matrix.setMat3( orient )

        return matrix
        
       
    @staticmethod
    def get_world_matrix( obj_name ):
        """

        """
        matrix_list = maya.cmds.xform( obj_name, query = True, matrix = True, worldSpace = True )
        rotatePivot = maya.cmds.xform( obj_name, query = True, ws = True, rotatePivot = True )

        matrix_list[ 12 ] = rotatePivot[ 0 ]
        matrix_list[ 13 ] = rotatePivot[ 1 ]
        matrix_list[ 14 ] = rotatePivot[ 2 ]

        #matrix = cgkit.cgtypes.mat4( matrix_list )
        #matrix = matrix.transpose( )

        return pm.datatypes.Matrix( matrix_list)    
        
        
    @staticmethod
    def set_world_matrix( obj_name, matrix, no_scale = False ):
        """

        """
        #if no_scale:
            #rot = matrix.getMat3( )
            #rot, scale = rot.decompose( )
            #matrix.setMat3( rot )

        #xform_list = matrix.toList( )
        pm.general.xform( obj_name, matrix = matrix, worldSpace = True )
        
        
    @staticmethod
    def get_third_axis(a1: Axis, a2: Axis):
        a3 = (a1 | a2) & ~Axis.REVERSE
        return Axis.ALL & ~a3
    
    
    @staticmethod
    def get_orthogonal_vectors( v1, v2, v3_dir = None ):
        """
        Given two vectors (v1 and v2) return two vectors (v2 and v3) that are orthogonal to v1.
        v3_dir can be used to ensure that v3 matches a desired direction
        """
        v3 = v1.cross( v2 ) #.normalize( )
        v3.normalize()
        
        #Make sure the up dir is aiming towards the desired up dir
        if v3_dir:
            if v3 * v3_dir < 0:
                v3 = -v3

        #Do a final adjustment of the r_vec to account for orthogonal consideration and a flipped Up vec
        scale = v2.length()
        v2 = v3.cross( v1 )
        v2.normalize( )
        v2 *= scale

        return (v2, v3)
 
##--methods-----------------------------------------------------------------------------------
                
                
    def set_forward_up_coordinates(self, forward: Axis,
                                   up: Axis):
        self._forward_axis = forward
        self._up_axis = up
        self._right_axis = self.get_third_axis(forward, up)
                
                
    def set_coordinates(self, foward: Axis,
                        up: Axis,
                        right: Axis):
        self._forward_axis = foward
        self._up_axis = up
        self._right_axis = right
        
    
    def get_matrix(self):
        return self.transform_node.getMatrix(worldSpace = self._space == Space.WORLD)
    
    
    def get_translation(self):
        if self.space == Space.WORLD:
            return self.transform_node.getTranslation('world')
        else:
            return self.transform_node.getTranslation('object')


    def _set_forward_up_right(self,
                               forward: pm.datatypes.Vector,
                               up: pm.datatypes.Vector,
                               right: pm.datatypes.Vector,
                               flip: Flip,
                               ignore_scale: bool):
        """
        Set's the transform_node's f, u, r to the input while ensuring proper right handedness.
        """
        
        MatrixUtils.ensure_right_handedness(forward, up, right, flip)
        
        if ignore_scale:
            forward.normalize()
            up.normalize()
            right.normalize()
        
        matrix = self.get_matrix()
        MatrixUtils.set_axis_vector(matrix, forward, self._forward_axis)
        MatrixUtils.set_axis_vector(matrix, up, self._up_axis)
        MatrixUtils.set_axis_vector(matrix, right, self._right_axis)
        
        self.transform_node.setMatrix(matrix, worldSpace = self.space == Space.WORLD)        
    
    
    def set_forward_up(self,
                       forward: pm.datatypes.Vector,
                       up: pm.datatypes.Vector,
                       priority: Direction = Direction.FORWARD, 
                       ignore_scale: bool = True):
        
        if (Direction.FORWARD in priority):
            up, right = self.get_orthogonal_vectors(forward, up)
            
        elif (Direction.UP in priority):
            forward, right = self.get_orthogonal_vectors(up, forward)
            
        else:
            pm.system.error('Type Error')
            
        self._set_forward_up_right(forward, up, right, Flip.RIGHT, ignore_scale)
        
        
    def set_forward_right(self,
                       forward: pm.datatypes.Vector,
                       right: pm.datatypes.Vector,
                       priority: Direction = Direction.FORWARD,
                       ignore_scale: bool = True):
        
        if (Direction.FORWARD in priority):
            right, up = self.get_orthogonal_vectors(forward, right)
            
        elif (Direction.RIGHT in priority):
            forward, up = self.get_orthogonal_vectors(right, forward)

        else:
            pm.system.error('Type Error')
            
        self._set_forward_up_right(forward, up, right, Flip.UP, ignore_scale)
        
        
    def set_up_right(self,
                       up: pm.datatypes.Vector,
                       right: pm.datatypes.Vector,
                       priority: Direction = Direction.UP,
                       ignore_scale: bool = True):
        
        if (Direction.UP in priority):
            right, forward = self.get_orthogonal_vectors(up, right)
            
        elif (Direction.RIGHT in priority):
            up, forward = self.get_orthogonal_vectors(right, up)
            
        else:
            pm.system.error('Type Error')
            
        self._set_forward_up_right(forward, up, right, Flip.FORWARD, ignore_scale)    
    
        
    



import pymel.core as pm
#iter = 10

def create_stretchy_spline():
    #make sure we have an ik handle in our selection
    selection = pm.general.ls(sl = True, type = 'ikHandle' )
    
    dm = pm.createNode('decomposeMatrix')
    npc = pm.createNode('nearestPointOnCurve')
    dm.outputTranslate >> npc.inPosition
    
    for handle in selection:
        inputCurve = pm.listConnections(handle.inCurve)
        if not inputCurve:
            continue
        
        inputCurve = inputCurve[0]
            
        #find the start and end joint
        startJoint = handle.getStartJoint()
        endEffector = handle.getEndEffector()

        endJoint = pm.listConnections(endEffector.tx)
        if endJoint:
            endJoint = endJoint[0]
        else:
            continue

        #build the network               
        inputCurve.worldSpace[0] >> npc.inputCurve
        currentJoint = endJoint          
        lastDistanceNode = None
        #count = 0
        
        while currentJoint != startJoint:
            currentJoint = currentJoint.getParent()
            #count += 0
            #if count > iter:
                #break
                
            #print(currentJoint)

            currentJoint.worldMatrix[0] >> dm.inputMatrix
            poc = pm.createNode('pointOnCurveInfo')
            inputCurve.worldSpace[0] >> poc.inputCurve
            
            #store the default parameter value for the joint
            #a quick connect then disconnect gets us what we need.
            npc.parameter >> poc.parameter
            npc.parameter // poc.parameter
            
            db = pm.createNode('distanceBetween')
            poc.position >> db.point1
            if lastDistanceNode:
                poc.position >> lastDistanceNode.point2
                divide = pm.createNode('multiplyDivide')

                lastDistanceNode.distance >> divide.input1X
                divide.input2X.set( lastDistanceNode.distance.get() )
                divide.operation.set(2)
                
                divide.outputX >> currentJoint.scaleX
                
            lastDistanceNode = db
             
    pm.delete(dm)
    pm.delete(npc)
    
    pm.select(selection, replace = True)
        
    
    

def Run():
    create_stretchy_spline()


import re
import enum
import maya.cmds
import pymel.core as pm


class Comp_Type(enum.Flag):
    VERTEX = enum.auto()
    EDGE = enum.auto()
    FACE = enum.auto()
    ALL = VERTEX | EDGE | FACE


def get_component_selection( comp_type: Comp_Type ):
    if not isinstance(comp_type, Comp_Type):
        maya.OpenMaya.MGlobal.displayError( "Component_type must be of type Comp_Type" )
        return
    
    abbr = ''
    if Comp_Type.VERTEX in comp_type:
        abbr = 'vtx'
    elif Comp_Type.EDGE in comp_type:
        abbr = 'e'
    else:
        abbr = 'f'
        

    expression = re.compile( r'{0}\[((?P<num1>\d+):(?P<num2>\d+)\]|(?P<num3>\d+)\])'.format( abbr ) )
    selection = maya.cmds.ls( sl = True )
    indices = []

    for item in selection:
        components = expression.search( item )

        if components:
            components = components.groupdict( )

            if components['num1'] is not None:
                for i in range( int( components['num1'] ), int( components['num2'] ) + 1 ):
                    indices.append( i )

            elif components['num3'] is not None:
                indices.append( int( components['num3'] ) )


    indices.sort()

    return indices



def  get_component_selections():
    results = {
        Comp_Type.VERTEX: get_component_selection(Comp_Type.VERTEX),
        Comp_Type.EDGE: get_component_selection(Comp_Type.EDGE),
        Comp_Type.FACE: get_component_selection(Comp_Type.FACE),
    }
    
    return results



def plot_percent_on_curve(curve : pm.nodetypes.NurbsCurve,
                          percent: float,
                          marker : pm.nodetypes.Transform):
    
    #I use a motion path constraint, because the percents are correct even
    #in cases where the curve isn't uniform.  E.g. one long span and two
    #short spans will put 50% 1/2 through span 2 when using PointOnCurve
    
    motionPath = pm.createNode('motionPath')
    motionPath.fractionMode.set(True)
    curve.worldSpace[0] >> motionPath.geometryPath
    motionPath.uValue.set(percent)
    motionPath.allCoordinates >> marker.translate
    motionPath.allCoordinates // marker.translate
    pm.general.delete(motionPath)
        
    
    
print("cg3dguru.utils.modeling loaded")


"""
Drag-n-drop this into a maya viewport to install packages

This is copied from:
https://github.com/Nathanieljla/cg3d-maya-core/blob/main/src/cg3dguru/utils/drop_installer.py

If you want to modify this script for your own purposes it's better
to pull a copy from the source location.
"""

__author__ = "Nathaniel Albright"
__email__ = "developer@3dcg.guru"

VERSION = (0, 9, 0)
__version__ = '.'.join(map(str, VERSION))



import os
import re
import time
import platform
import sys
import webbrowser
import base64
#import math
#from datetime import datetime, timedelta
import glob
import tempfile
import shutil
import sys
import subprocess
from os.path import expanduser
import zipfile
from functools import partial
import site

try:
    #python3
    from urllib.request import urlopen
except:
    #python2
    from urllib import urlopen

try:
    #python2
    reload
except:
    #python3
    from importlib import reload

try:
    import maya.utils
    import maya.cmds
    from maya import OpenMayaUI as omui
    
    from PySide2.QtCore import *
    from PySide2.QtWidgets import *
    from PySide2.QtGui import *
    from shiboken2 import wrapInstance
    MAYA_RUNNING = True
except ImportError:
    MAYA_RUNNING = False
    

RESOURCES = None
    
class Platforms(object):
    OSX = 0,
    LINUX = 1,
    WINDOWS = 2
    
    @staticmethod
    def get_name(enum_value):
        if enum_value == Platforms.OSX:
            return 'osx'
        elif enum_value == Platforms.LINUX:
            return 'linux'
        else:
            return 'windows'
        
        
        
class Commandline(object):
    @staticmethod        
    def get_python_version():
        """Get the running version of python as a tuple of 3 ints"""
        pmax, pmin, patch =  sys.version.split(' ')[0].split('.')
        return( int(pmax), int(pmin), int(patch))
    
    
    @staticmethod
    def get_platform():
        result = platform.platform().lower()
        if 'darwin' in result:
            return Platforms.OSX
        elif 'linux' in result:
            return Platforms.LINUX
        elif 'window' in result:
            return Platforms.WINDOWS
        else:
            raise ValueError('Unknown Platform Type:{0}'.format(result))
    
    
    @staticmethod    
    def run_shell_command(cmd, description):
        #NOTE: don't use subprocess.check_output(cmd), because in python 3.6+ this error's with a 120 code.
        print('\n{0}'.format(description))
        print('Calling shell command: {0}'.format(cmd))

        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, stderr = proc.communicate()
        stdout = stdout.decode()
        stderr = stderr.decode()
        
        print(stdout)
        print(stderr)
        if proc.returncode:
            raise Exception('Command Failed:\nreturn code:{0}\nstderr:\n{1}\n'.format(proc.returncode, stderr))
        
        return(stdout, stderr)
    
        
    @staticmethod
    def get_python_paths():
        """Returns maya's python path and location of a global pip
        
        Note: The pip path might not exist on the system.
        """
        python_path = ''
        pip_path = ''
        pmax, pmin, patch = Commandline.get_python_version()
        platform = Commandline.get_platform()
        
        version_str = '{0}.{1}'.format(pmax, pmin)
        if platform == Platforms.WINDOWS:
            python_path = os.path.join(os.getenv('MAYA_LOCATION'), 'bin', 'mayapy.exe')
            if pmax > 2:
                #python3 pip path
                pip_path = os.path.join(os.getenv('APPDATA'), 'Python', 'Python{0}{1}'.format(pmax, pmin), 'Scripts', 'pip{0}.exe'.format(version_str))
            else:
                #python2 pip path
                pip_path = os.path.join(os.getenv('APPDATA'), 'Python', 'Scripts', 'pip{0}.exe'.format(version_str))

        elif platform == Platforms.OSX:
            python_path = '/usr/bin/python'
            pip_path = os.path.join( expanduser('~'), 'Library', 'Python', version_str, 'bin', 'pip{0}'.format(version_str) )
     
        elif platform == Platforms.LINUX:
            python_path = os.path.join(os.getenv('MAYA_LOCATION'), 'bin', 'mayapy')
            pip_path = os.path.join( expanduser('~'), '.local', 'bin', 'pip{0}'.format(version_str) )
             
        return (python_path, pip_path)
    
    
    
    @staticmethod
    def get_command_string():
        """returns a commandline string for launching pip commands
        
        If the end-user is on linux then is sounds like calling pip from Mayapy
        can cause dependencies issues when using a default python install.
        So if the user is on osX or windows OR they're on linux and don't
        have python installed, then we'll use "mayapy -m pip" else we'll
        use the pipX.exe to run our commands.        
        """
        python_path, pip_path = Commandline.get_python_paths()
        platform = Commandline.get_platform()        

        command = '{0}&-m&pip'.format(python_path)
        global_pip = False
        if platform == Platforms.LINUX:
            try:
                #I don't use "python" here, because on windows that opens the MS store vs. erroring.
                #No clue what it might do on linux
                Commandline.run_shell_command(['py'], 'Checking python install')
                command = pip_path
                global_pip = True
            except:
                #Python isn't installed on linux, so the default command is good
                pass
            
        return (command,  global_pip)

                
    @staticmethod
    def pip_install(package, pip_args = [], *args, **kwargs):
        pip_command, global_pip = Commandline.get_command_string()
        cmd_str = ('{0}&install&{1}').format(pip_command, package)
        args = cmd_str.split('&') + pip_args
        stdout, stderr = Commandline.run_shell_command(args, 'PIP:Installing Package')
        
        return stdout
    
    
    @staticmethod
    def pip_uninstall(package, pip_args = [], *args, **kwargs):
        pip_command, global_pip = Commandline.get_command_string()
        cmd_str = ('{0}&uninstall&{1}').format(pip_command, package)
        args = cmd_str.split('&') + pip_args
        stdout, stderr = Commandline.run_shell_command(args, 'PIP:Installing Package')
        
        return stdout
    

    @staticmethod
    def pip_list(pip_args = [], *args, **kwargs):
        pip_command, global_pip  = Commandline.get_command_string()
        cmd_str = ('{0}&list').format(pip_command)
        args = cmd_str.split('&') + pip_args
        stdout, stderr = Commandline.run_shell_command(args, 'PIP:Listing Packages')
        
        return stdout    


    @staticmethod
    def pip_show(repo_name, pip_args = [], *args, **kwargs):
        pip_command, global_pip  = Commandline.get_command_string()
        cmd_str = ('{0}&show&{1}').format(pip_command, repo_name)
        args = cmd_str.split('&') + pip_args
        stdout, stderr = Commandline.run_shell_command(args, 'PIP:Show Package Info')
        
        return stdout        
        

        
class ModuleDefinition(object):
    """A .mod file can have multiple entries.  Each definition equates to one entry"""
    
    MODULE_EXPRESSION = r"(?P<action>\+|\-)\s*(MAYAVERSION:(?P<maya_version>\d{4}))?\s*(PLATFORM:(?P<platform>\w+))?\s*(?P<module_name>\w+)\s*(?P<module_version>\d+\.?\d*.?\d*)\s+(?P<module_path>.*)\n(?P<defines>(?P<define>.+(\n?))+)?"
        
    def __init__(self, module_name, module_version,
                 maya_version = '', platform = '',
                 action = '+', module_path = '',
                 defines = [],
                 *args, **kwargs):
        
        self.action = action
        self.module_name = module_name
        self.module_version = module_version
        
        self.module_path = r'.\{0}'.format(self.module_name)
        if module_path:
            self.module_path = module_path

        self.maya_version = maya_version
        if self.maya_version is None:
            self.maya_version = ''
        
        self.platform = platform
        if self.platform is None:
            self.platform = ''
        
        self.defines = defines
        if not self.defines:
            self.defines = []
        
    def __str__(self):
        return_string = '{0} '.format(self.action)
        if self.maya_version:
            return_string += 'MAYAVERSION:{0} '.format(self.maya_version)
            
        if self.platform:
            return_string += 'PLATFORM:{0} '.format(self.platform)
            
        return_string += '{0} {1} {2}\n'.format(self.module_name, self.module_version, self.module_path)
        for define in self.defines:
            if define:
                return_string += '{0}\n'.format(define.rstrip('\n'))
         
        return_string += '\n'    
        return return_string



class ModuleManager(QThread):
    """Used to edit .mod files quickly and easily."""
    
    def __init__(self, module_name, module_version, package_name='',
                 include_site_packages = False):
        
        QThread.__init__(self)
        self.install_succeeded = False
        
        self._module_definitions = []
        self.module_name = module_name
        self.module_version = module_version
        
        self.package_name = package_name
        if not self.package_name:
            self.package_name = self.module_name
        
        self.maya_version = self.get_app_version()
        self.platform = Commandline.get_platform()
        
        self.max, self.min, self.patch = Commandline.get_python_version()
        
        #common locations
        self._version_specific = self.is_version_specific()  
        self.app_dir = os.getenv('MAYA_APP_DIR')
        self.install_root = self.get_install_root()
        self.relative_module_path = self.get_relative_module_path()
        self.module_path = self.get_module_path()
        self.icons_path = self.get_icon_path()
        self.presets_path = self.get_presets_path()
        self.scripts_path = self.get_scripts_path()
        self.plugins_path = self.get_plugins_path()
        
        self.site_packages_path = self.get_site_package_path()
        if not include_site_packages:
            self.site_packages_path = ''
            
        self.package_install_path = self.get_package_install_path()
        
        #Non-Maya python and pip paths are needed for installing on linux (and OsX?)
        self.python_path, self.pip_path = Commandline.get_python_paths()
        self.command_string, self.uses_global_pip = Commandline.get_command_string()
     
    
    def __del__(self):
        #TODO: Determine why I put a wait on this fuction 
        self.wait()
 
       
    @staticmethod
    def get_app_version():
        """What version of Maya is this?"""
        return int(str(maya.cmds.about(apiVersion=True))[:4])
    
    
    @staticmethod
    def get_platform_string(platform):
        """Convert the current Platform value to a Module string"""
        if platform is Platforms.OSX:
            return 'mac'
        elif platform is Platforms.LINUX:
            return 'linux'
        else:
            return 'win64'
    
    
    @staticmethod
    def make_folder(folder_path):
        print(folder_path)
        
        if not os.path.exists(folder_path):
            os.makedirs(folder_path)


    @staticmethod
    def get_ui_parent():
        return wrapInstance( int(omui.MQtUtil.mainWindow()), QMainWindow )      
   
    
    @staticmethod
    def package_installed(package_name):
        """returns True if the repo is already on the system"""
        
        return Commandline.pip_list().find(package_name) != -1
    

    @staticmethod
    def package_outdated(package_name):
        """Check to see if a local package is outdated
        
        Checks to see the local pacakge is out-of-date.  This will always
        be true with remote packages that are from Git, but will be accurate
        with packages registered on PyPi. Since package_outdated()
        assumes the package exists before checking make sure you you first
        check the existance of the package with package_installed() before
        checking the outdated status.
        
        Returns:
        --------
        bool
            True if missing or outdated, else False
        """
        #TODO: get version checking to work with git packages.
        #https://stackoverflow.com/questions/11560056/pip-freeze-does-not-show-repository-paths-for-requirements-file
        #https://github.com/pypa/pip/issues/609
        #it looks like we'd have to install packages with pip -e for this to work,
        #but then we can't install to a target dir. I'm getting errors about
        #trying to install in maya/src, but --user doesn't work either.

        #I'm using --uptodate here, because both --uptodate and --outdated
        #will be missing the package if the pacakage isn't registered with PyPi
        #so -uptodate is easier to verify with than -o with remote package that
        # might or might not be registered with PyPi
        
              
        result = Commandline.pip_list(pip_args =['--uptodate'])
        outdated = result.find(package_name) == -1
        if outdated:
            return True
        else:
            return False
    

    def install_remote_package(self, package_name = '', to_module = True):
        if not package_name:
            package_name = self.get_remote_package()
        
        #https://stackoverflow.com/questions/39365080/pip-install-editable-with-a-vcs-url
        #github = r'https://github.com/Nathanieljla/fSpy-Maya.git'
        
        pip_args = []
        if to_module:
            pip_args = [
                #r'--user', 
                #r'--editable=git+{0}#egg={1}'.format(github, self.repo_name), 
                r'--target={0}'.format(self.scripts_path), 
            ]
        Commandline.pip_install(package_name, pip_args)
    
    
    def get_remote_package(self):
        """returns the github or PyPi name needed for installing"""
        maya.cmds.error( "No Package name/github path defined.  User needs to override Module_manager.get_remote_package()" )

        
    def __ensure_pip_exists(self):
        """Make sure OS level pip is installed
        
        This is written to work with all platforms, but
        I've updated this to only run when we're on linux
        because it sounds like that's the only time it's needed
        """
        
        if not self.uses_global_pip:
            print("Using Maya's PIP")
            return
        
        if os.path.exists(self.pip_path):
            print('Global PIP found')
            return
        
        tmpdir = tempfile.mkdtemp()
        get_pip_path = os.path.join(tmpdir, 'get-pip.py')
        print(get_pip_path)
        
        if self.platform == Platforms.OSX:
            #cmd = 'curl https://bootstrap.pypa.io/pip/{0}/get-pip.py -o {1}'.format(pip_folder, pip_installer).split(' ')
            cmd = 'curl https://bootstrap.pypa.io/pip/get-pip.py -o {0}'.format(get_pip_path).split(' ')
            Commandline.run_shell_command(cmd, 'get-pip')

        else:
            # this should be using secure https, but we should be fine for now
            # as we are only reading data, but might be a possible mid attack
            #response = urlopen('https://bootstrap.pypa.io/pip/{0}/get-pip.py'.format(pip_folder))
            response = urlopen('https://bootstrap.pypa.io/pip/get-pip.py')
            data = response.read()
            
            with open(get_pip_path, 'wb') as f:
                f.write(data)
                
        # Install pip
        # On Linux installing pip with Maya Python creates unwanted dependencies to Mayas Python version, so pip might not work 
        # outside of Maya Python anymore. So lets install pip with the os python version. 
        filepath, filename = os.path.split(get_pip_path)
        #is this an insert, so this pip is found before any other ones?
        sys.path.insert(0, filepath)
        
        
        if self.platform == Platforms.OSX or self.platform == Platforms.LINUX:
            python_str = 'python{0}.{1}'.format(self.max, self.min)
        else:
            python_str = self.python_path
            
        cmd = '{0}&{1}&--user&pip'.format(python_str, get_pip_path).split('&')
        Commandline.run_shell_command(cmd, 'install pip')
        
        print('Global PIP is ready for use!')
        
        
    def is_version_specific(self):
        """Is this install for a specific version of Maya?
        
        Some modules might have specific code for different versions of Maya.
        For example if Maya is running Python 3 versus. 2. get_relative_module_path()
        returns a different result when this True vs.False unless overridden by
        the user.
        
        Returns:
        --------
        bool
            False
        """        
        
        return False
     
    def get_install_root(self):
        """Where should the module's folder and defintion install?
        
        Maya has specific locations it looks for module defintitions os.getenv('MAYA_APP_DIR')
        For windows this is "documents/maya/modules" or "documents/maya/mayaVersion/modules"
        However 'userSetup' files can define alternative locations, which is
        good for shared modules in a production environment.
        
        Returns:
        --------
        str
            os.path.join(self.app_dir, 'modules')
        """        
        return os.path.join(self.app_dir, 'modules')
    
    def get_relative_module_path(self):
        """What's the module folder path from the install root?
        
        From the install location we can create a series of folder to reach
        the base of our module.  This is where Maya will look for the
        'plug-ins', 'scripts', 'icons', and 'presets' dir.  At a minimum
        you should return the name of your module. The default implementation
        create as a path of 'module-name'/platforms/maya-version/platform-name/x64
        when is_version_specific() returns True
        
        Returns:
        str
            self.module_name when is_version_specific() is False
        
        """
        root = self.module_name
        if self._version_specific:
            root = os.path.join(self.module_name, 'platforms', str(self.maya_version),
                                Platforms.get_name(self.platform),'x64')  
        return root
    
    def get_module_path(self):
        return os.path.join(self.install_root, self.relative_module_path)
    
    def get_icon_path(self):
        return os.path.join(self.module_path, 'icons')
    
    def get_presets_path(self):
        return os.path.join(self.module_path, 'presets')
    
    def get_scripts_path(self):
        return os.path.join(self.module_path, 'scripts')
    
    def get_plugins_path(self):
        return os.path.join(self.module_path, 'plug-ins')
    
    def get_site_package_path(self):
        return os.path.join(self.scripts_path, 'site-packages')
    
    def get_package_install_path(self):
        return os.path.join(self.scripts_path, self.module_name)
    
  
    def read_module_definitions(self, path):
        self._module_definitions = []
        if (os.path.exists(path)):
            file = open(path, 'r')
            text = file.read()
            file.close()
          
            for result in re.finditer(ModuleDefinition.MODULE_EXPRESSION, text):
                resultDict = result.groupdict()
                if resultDict['defines']:
                    resultDict['defines'] = resultDict['defines'].split("\n")
                    
                definition = ModuleDefinition(**resultDict)
                self._module_definitions.append(definition)
      
                        
    def write_module_definitions(self, path):
        file = open(path, 'w')
        for entry in self._module_definitions:
            file.write(str(entry))
        
        file.close()

                           
    def __get_definitions(self, search_list, key, value):
        results = []
        for item in search_list:
            if item.__dict__[key] == value:
                results.append(item)
                
        return results
        
          
    def _get_definitions(self, *args, **kwargs):
        result_list = self._module_definitions
        for i in kwargs:
            result_list = self.__get_definitions(result_list, i, kwargs[i])
        return result_list
    
    
    def remove_definitions(self, *args, **kwargs):
        """
        removes all definitions that match the input argument values
        example : module_manager_instance.remove_definitions(module_name='generic', platform='win', maya_version='2023')
        
        Returns:
        --------
        list
            the results that were removed from the manager.
        
        """ 
        results = self._get_definitions(**kwargs)
        for result in results:
            self._module_definitions.pop(self._module_definitions.index(result))
            
        return results
    
    
    def remove_definition(self, entry):
        self.remove_definitions(module_name=entry.module_name,
                                platform=entry.platform, maya_version=entry.maya_version)
    
    def add_definition(self, definition):
        """

        """
        #TODO: Add some checks to make sure the definition doesn't conflict with an existing definition
        self._module_definitions.append(definition)
        
   
    def run(self):
        """this starts the QThread"""
        try:
            self.install_succeeded = self.install()
        except Exception as e:
            self.install_succeeded = False
            print('Install Failed!!\n{0}'.format(e))
            
                 
    def get_definition_entry(self):
        """Converts this class into a module_defintion
        
        Returns:
        --------
        Module_definition
            The module defintion that represents the data of the Module_manager
        
        """
        maya_version = str(self.maya_version)
        relative_path = '.\{0}'.format(self.relative_module_path)        
        platform_name =  self.get_platform_string(Commandline.get_platform())
        
        if not self._version_specific:
            maya_version = ''
            platform_name = ''
            
        defines = []
        if self.site_packages_path:
            site_path =  'PYTHONPATH+:={0}'.format(self.site_packages_path.split(self.module_path)[1])
            defines = [site_path]
        
        module_definition = ModuleDefinition(self.module_name, self.module_version,
                                                             maya_version=maya_version, platform=platform_name, 
                                                             module_path=relative_path,
                                                             defines=defines)
        return module_definition
     
             
    def update_module_definition(self, filename):
        """remove old defintions and adds the current defintion to the .mod
        
        Returns:
        --------
        bool
            True if the update was successful else False        
        """
        new_entry = self.get_definition_entry()
        self.remove_definition(new_entry) #removes any old entries that might match before adding the new one
        self.add_definition(new_entry)  
        try:
            self.write_module_definitions(filename)
        except IOError:
            return False
        
        return True
        

    def pre_install(self):
        """Called before install() to do any sanity checks and prep
        
        This function attempts to create the required install folders
        and update/add the .mod file. Sub-class should call this function
        when overriding

        Returns:
        --------
        bool
            true if the install can continue
        """
        try:
            self.__ensure_pip_exists()

        except Exception as e:
            print('failed to setup global pip {0}'.format(e))
            return False
        
        try:          
            self.make_folder(self.module_path)       
            self.make_folder(self.icons_path)
            self.make_folder(self.presets_path)
            self.make_folder(self.scripts_path)
            self.make_folder(self.plugins_path)
            
            if self.site_packages_path:
                self.make_folder(self.site_packages_path)
        except OSError:
            return False

        filename = os.path.join(self.install_root, (self.module_name + '.mod'))
        self.read_module_definitions(filename)
              
        return self.update_module_definition(filename)
    

    def install(self):
        """The main install function users should override"""        
        installed = False
        if not self.package_installed(self.package_name):

            try:
                self.install_remote_package()
                installed = True
            except:
                pass
            
        return installed
    
    
    def post_install(self):
        """Used after install() to do any clean-up

        """  
        print('post install')
        if self.install_succeeded:
            if self.scripts_path not in sys.path:
                sys.path.append(self.scripts_path)
                print('Add scripts path [{}] to system paths'.format(self.scripts_path))
            else:
                print('scripts path in system paths')
                
                
    def install_pymel(self):
        """Installs pymel to a common Maya location"""
        if not self.package_installed('pymel'):
            Commandline.pip_install('pymel')
        
    
    
##-----begin UI----##
    
class IconButton(QPushButton):
    def __init__(self, text, highlight=False, icon=None, success=False, *args, **kwargs):
        super(IconButton, self).__init__(QIcon(icon), text, *args, **kwargs)

        self.icon = icon
        self.highlight = highlight
        self.success = success
        self.setMinimumHeight(34)
        self.setCursor(QCursor(Qt.PointingHandCursor))
        if self.highlight:
            font = self.font()
            font.setPointSize(14)
            font.setBold(True)
            self.setFont(font)

        if self.success:
            font = self.font()
            font.setPointSize(14)
            font.setBold(True)
            self.setFont(font)

        if self.icon:
            self.setIconSize(QSize(22, 22))
            self.setIcon(QIcon(self.AlphaImage()))

    def AlphaImage(self):
        if self.highlight and not self.success:
            AlphaImage = QPixmap(self.icon)
            painter = QPainter(AlphaImage)

            painter.setCompositionMode(QPainter.CompositionMode_SourceIn)
            painter.fillRect(AlphaImage.rect(), '182828')

            return AlphaImage

        else:
            return QPixmap(self.icon)
        
     

class Resources(object):
    preloaderAnimBase64 = ''' '''
    BYTE_GIF = QByteArray.fromBase64( preloaderAnimBase64.encode() )
    
    company_base64 = None
    install_base64 = '''iVBORw0KGgoAAAANSUhEUgAAAFgAAABYCAYAAABxlTA0AAAABHNCSVQICAgIfAhkiAAAAAlwSFlzAAALEwAACxMBAJqcGAAAA2NJREFUeJztnN9RGzEQh3+ryTvuIHQAqSCOzn4OJUAFoYO4BKeCpAR4PkvjdGA6oATTgDYvx4xjwv3x7Z7kyX5v3N5I6w8h6eQ9AMMwDMMwDMMwDMMwDMMwDON/gHIn0JcY4+fDn733v3PlMoTiBdd1feGcewYwO7zOzNvFYvElT1b9cbkT6MI5t8aRXAAgonkI4XuGlAZRvGAANy2x+8myOJFzEPxm9PaMFcE5CD5rTLAyJlgZE6yMCVbGBCtjgpX5oNHoZrP5hr8fEPbOuZX3/kmjv6HEGK9SSuvDa865tff+Ubov8bOIGONXZn74R+g5pXS9XC5fhrQXQuC2eFVVgz5Dc7axA3B5FNoT0Vx6EIhPESml1TuhSyLa1nV9Id1nX+q6viCiLd7KBYDZ8aiWQFwwEV23xXJJfpXbkd9cut/JF7kckvvI1UJcMDNvu+6ZUvIQuX1yH4q4YOfcPYB9131TSB44cvdN7qKIC/bePzVzWVbJQ+Vq7CAApTk4t+RS5AKKi1wuySXJBZR3EVNLLk0uMME2bSrJJcoFJtoHa0suVS4wcV1EjPGq2Wt2flnJzDtmnjvnWn8pKaVZqXKBDIUnQyV3ietzT8PkcoFMlT1DJAuRRS6QsXRqQsnZ5AKZa9MmkJxVLlBA8Z+i5OxygQIEAyqSi5ALFCIYEJVcjFygIMGAiOSi5AKFCQZGSS5OLiAoWLLE/wTJo+VqvaIwWrBWif8AyaPkar+iMPqwR6vEv+cB0eiRq/2KgsRpmlqJfyP59p2w1Jyr+oqChGDVEn/v/SMRXTPz7vUaM29TSpdCC5pq/iq1adI0Ij/lzuMU1A/cY4xX2n2cyvHOQYPRgruKNVJK87F9aJFSap1jJQpRRk8RRPTcEV9tNhsw86+hlZVahBA+MvOaiNoWOBDRri3eh9GCmXnbstIDwIyI1kS0DiGM7U4Mou5HgKYScxQSU8QDenyZeYbsJQqyRwtu/uzF62pzw8wriXZEdhEppfXhPvXcYebdYrH4IdGWiODlcvnCzHON8s+pYeYtM8+l2hM/rgwh/ARwK93uRPyqqupOskHxB42qqu6Y+RZntvAx8720XEDxwL0pZ7oBcNOcihX3rweaKe2hpD26YRiGYRiGYRiGYRiGYRiGYRiGYRjD+AN6etoeN+hk/AAAAABJRU5ErkJggg=='''
    close_base64 = '''iVBORw0KGgoAAAANSUhEUgAAAFgAAABYCAYAAABxlTA0AAAABHNCSVQICAgIfAhkiAAAAAlwSFlzAAALEwAACxMBAJqcGAAAA7lJREFUeJztm8tx2zAQhv9lA3YHdiqwO4iGpM5xB6Y7UAdRKohSQeQO5DMfo3Qgd6B0IDfAzSHgjOwhJQDCgvTMfleCXPIbEI8FACiKoiiKoiiKoiiKoiiKoiiKoiiKoijKECT14LqubwAUzDwDACLatW27nM/nb1IxbWia5o6ZF8x8a95rT0SrNE1fJeKJCK7r+ieARc+ldZZlTxIxbSjL8oqItkR0//EaEd1LSA4u2NSQ3Yki67ZtF7Frcl3XN8y86ZNr2GdZ9iV03CT0A5n54UyRgoi2ZVlehY49RNM0dwB2J+QCwK1EbAnBh3NliOg+lmTzR20BXEvH6iO44CRJtjblYkh2kXumWfMmuOA0TV+ZeWNTlojukyTZm184KFVVPRpptjV3HfodAAHBAMDMhUONuGbmbUjJVVU9EtHa4ZZ1nue/QsU/RkTwfD5/Y+aZ+T1tCCbZQ+5KcugoNtHoqOv6N4DCsviBmRd5nj97xhoaf/fCzIVvLFvEBQPOkr0+PEYMH6IIBmQFTFUuEFEw4NU+LrMs+zF00Ux910R0bnLTcSCimVTeoY+oggG/Hr6vEzqVVxggulxgBMHA5ZI/i1xgJMEA0DTNVzMhsZ4ItG27SJLk+kzS5h3MvCOihyzL/vq/rT+jCQbc8wRG1q1LeWaejZmDHlUwIJeMmYJcQGgm50Kapq9ENAuZbGHmzRTkAhOowR0eHdcQo66afGQygoEgkiclF5hAE3GMR5LoGNGkjS+TEgz8l0xEW8fbDh73RGFSTQTgnlc4JmaOwZZJCb5EbsfUJE9CcFmWV0mSrHCh3A6TUxZZoXBldMEBh2cfmcSIYtROzidp4/D4wjQ5ozKaYFe5Jg8xM7lfW9GjSx6lifBJ8hxPfS+9PybRBYeS81kkR20iPKRsh6R0SSIAe5tnxdyu9S5urEBmFWMFhwS7zSjAo6PcE9FDrNWNKIJDrcMNMeUlJHHB0nI7pipZVHBd198BLG3Lh5jmuu4kIqIiTdOXS2KeQvKMxmibQaa0EUVEsGuzIPGBrpJNxxe8JosM04hoaVn0QEQzidqTZdkTM1tvBGzbdhn6HQABwVVVPcLuvEPXyfwJ/Q4deZ7/YubCpqxAsgmAgGCzb+Ec0YZJeZ4/20qWQELwyeMDXdIm5jamPM+fLZJEe4nYImc0AKz6rnX5gDH2iKVp+mKm1n2S9w47NJ0QG6Y1TfOtbdvjTmbNzJuxN4P0HPHdj3EwUlEURVEURVEURVEURVEURVEURVEURVGAf9l4XANGvwF5AAAAAElFTkSuQmCC'''

    def __init__(self):
        super(Resources, self).__init__()        
        self.installer = None
        
    def __del__(self):
        if self.installer:
            self.installer.deleteLater()
             
    @staticmethod
    def base64_to_QPixmap(base64Image):
        pixmap = QPixmap()
        byte_array =  QByteArray.fromBase64( base64Image.encode() )
        pixmap.loadFromData(byte_array)
        
        return QPixmap(pixmap)
    
    @staticmethod
    def qPixmap_to_base64(pixmap, extension):
        #https://doc.qt.io/qtforpython/index.html
        #https://forum.qt.io/topic/85064/qbytearray-to-string/2
        image = pixmap.toImage()
        byteArray = QByteArray()
        buffer = QBuffer(byteArray)
        image.save(buffer, "png")
        base64 = byteArray.toBase64().data()
        result = str(base64, encoding='utf-8')
            
        return result
            
    @staticmethod
    def file_to_base64(file_path):
        extension = os.path.basename(file_path).split('.')[-1]
        pixmap = QPixmap()
        if pixmap.load(file_path):
            return Resources.qPixmap_to_base64(pixmap, extension)
        else:
            return ''
        
    @staticmethod
    def print_file_string(image_path):
        image_string =  Resources.file_to_base64(image_path)
        if image_string:
            print ("Use the next line for the Resources.company_base64")
            print(image_string)
  
      
    def set_installer(self, installer):
        self.installer = installer
    
    
    @property
    def close_icon(self):        
        return self.base64_to_QPixmap(Resources.close_base64)
    
    @property
    def install_icon(self):        
        return self.base64_to_QPixmap(Resources.install_base64)
              
    @property
    def company_icon(self):
        result = None
        if Resources.company_base64:
            result = self.base64_to_QPixmap(Resources.company_base64)
        
        return result
    
    @company_icon.setter
    def company_icon(self, value):
        Resources.company_base64 = value
      
    
class InstallerUi(QWidget):
    def __init__(self, name, module_manager, background_color = '',
                 company_logo_size = [64, 64],
                 launch_message='', 
                 installing_message = 'Installing, please wait ...',
                 failed_message='Install Failed!',
                 success_message="Install Completed Successfully!",
                 post_error_messsage='Install Successful. Clean-up errored.  See output.', 
                 *args, **kwargs
                 ):
        
        parent = module_manager.get_ui_parent()
        super(InstallerUi, self).__init__(parent=parent, *args, **kwargs)
        
        self.name = name
        self.module_manager = module_manager
        
        self.launch_message = launch_message
        self.installing_message = installing_message
        self.failed_message = failed_message
        self.success_message = success_message
        self.post_error_messsage =  post_error_messsage
        
        self.create_layout(background_color, company_logo_size)
        self.set_default_size(name)
        self.install_button.clicked.connect(self.on_install)
        self.close_button.clicked.connect(self.on_close)
        #

    def set_default_size(self, name):
        self.animated_gif.hide()         
        self.setObjectName(name)
        self.setWindowTitle(name)
        self.setWindowModality(Qt.ApplicationModal)
        self.setWindowFlags(Qt.FramelessWindowHint)
        self.setWindowFlags(Qt.Tool)
        self.setFixedSize(self.layout().minimumSize())
        self.close_button.hide()
        
        size = self.layout().minimumSize()
        width = size.width()
        height = size.height()
        desktop = QApplication.desktop()
        screenNumber = desktop.screenNumber(QCursor.pos())
        screenRect = desktop.screenGeometry(screenNumber)
        widthCenter = (screenRect.width() / 2) - (width / 2)
        heightCenter = (screenRect.height() / 2) - (height / 2)        
        self.setGeometry(QRect(widthCenter, heightCenter, width, height))
        
        
    def create_layout(self, background_color, company_logo_size):
        #background color
        if background_color:
            palette = self.palette()
            palette.setColor(self.backgroundRole(), background_color)
            self.setPalette(palette)             
        
        ##-----create all our ui elements THEN arrange them----##
        logo = None

        if RESOURCES.company_icon is not None:
            logo = QLabel()
            smallLogo = RESOURCES.company_icon.scaled(company_logo_size[0], company_logo_size[1], Qt.KeepAspectRatio, Qt.SmoothTransformation)
    
            logo.setPixmap(smallLogo)
            logo.setAlignment(Qt.AlignCenter | Qt.AlignCenter)
            logo.setMargin(15)

        self.install_button = IconButton('Install', highlight=True) #, icon=RESOURCES.install_icon)
        self.install_button.setMinimumHeight(42)
        self.close_button = IconButton(' Close', icon=RESOURCES.close_icon)
        self.close_button.setMinimumHeight(42)

        self.message_label = QLabel()
        self.message_label.setText(self.launch_message)
        self.message_label.show()
        
        self.movie = QMovie()
        self.device = QBuffer(Resources.BYTE_GIF)
        self.movie.setDevice(self.device)
        self.animated_gif = QLabel()

        self.animated_gif.setMovie(self.movie)
        self.animated_gif.setMaximumHeight(24)
        self.animated_gif.setAlignment(Qt.AlignLeft | Qt.AlignVCenter)
        self.animated_gif.setScaledContents(True)
        self.animated_gif.setMaximumWidth(24)
        outer = QVBoxLayout()
        self.setLayout(outer)
        if logo:
            outer.addWidget(logo, 0)
            
        message_layout = QVBoxLayout()
        message_layout.addWidget(self.message_label, 1)
        message_layout.setAlignment(Qt.AlignCenter)
        outer.addLayout(message_layout)
        
        button_layout = QHBoxLayout()
        button_layout.addStretch()
        button_layout.addWidget(self.install_button, 0)
        button_layout.addWidget(self.close_button, 0)
        button_layout.addWidget(self.animated_gif, 0)
        button_layout.addStretch()
        button_layout.setAlignment(Qt.AlignCenter)

        outer.addLayout(button_layout)
        self.layout()
                
    def on_install(self):
        self.install_button.hide()
        #self.movie.start() #I'm thinking this causes maya to crash when debugging in WING
        self.animated_gif.show()
        self.message_label.setText(self.installing_message)
        self.message_label.show()
        
        if self.module_manager.pre_install():
            self.connect(self.module_manager, SIGNAL('finished()'), self.done)
            self.module_manager.start()
        else:
            self.close_button.show()
            self.animated_gif.hide()
            self.message_label.setText(self.failed_message)
        
    
    def done(self):
        self.close_button.show()
        self.animated_gif.hide()
        if self.module_manager.install_succeeded:
            self.message_label.setText(self.success_message)
        else:
            self.message_label.setText(self.failed_message)
        
        no_errors = self.module_manager.post_install()
        if not no_errors:
            self.message_label.setText(self.post_error_messsage)
    
        
    def on_close(self):
        self.close()
        
    def clean(self):
        self.movie.stop()
        if self.device.isOpen():
            self.device.close()

    def closeEvent(self, event):
        self.clean()
        
        
        
class MyInstaller(ModuleManager):
    def __init__(self, *args, **kwargs):
        super(MyInstaller, self).__init__(*args, **kwargs)

        
    def get_remote_package(self):
        """Users must override this with pypi name or github address"""
        return super(MyInstaller, self).get_remote_package()
              
        
    
def main():
    if MAYA_RUNNING:
        MODULE_NAME = 'module name'
        MODULE_VERSION = 1.0
        PACKAGE_NAME = MODULE_NAME
        
        manager = MyInstaller(MODULE_NAME, MODULE_VERSION, package_name = PACKAGE_NAME)
        
        
        Resources.company_base64 = '''''' #must be surrounded by three '.  Uses Resources.print_file_string to generate a string from an iFXf00cV114ukwm9xUSVHtiM
        logo_size = [64, 64]
        background_color = '' #hex string

        global RESOURCES
        RESOURCES = Resources()
        window_name =  manager.module_name.replace("_", " ")
        installer = InstallerUi(window_name, manager, background_color=background_color, company_logo_size=logo_size)
        RESOURCES.set_installer(installer)

        installer.show()

           
def onMayaDroppedPythonFile(*args):
    main()
    

def run():
    """
    Run is a function used by WingIDE to execute code after telling Maya to import the module
    """    
    main()
    

if __name__ == "__main__":
    main()